import json
import os
import pickle
from collections.abc import Callable, Sequence
from enum import IntEnum, auto
from typing import Any
from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient
from pydantic import BaseModel, Field
from pytest_subtests import SubTests
from typeguard import suppress_type_checks

from aviary.core import (
    INVALID_TOOL_NAME,
    DummyEnv,
    Environment,
    FunctionInfo,
    Message,
    Tool,
    ToolCall,
    ToolRequestMessage,
    argref_by_name,
)
from aviary.tools.server import make_tool_server


def simple() -> None:
    """Doing nothing may be better than doing something."""


def intuitive_arg(x: str) -> float:  # type: ignore[empty-body]
    """Cast the input argument x to a float."""


class StubState(BaseModel):
    """Stub model docstring."""

    defaulted_int: int = Field(default=1, description="A description of the int.")
    required_str: str = Field(description="A description of the str.")


class StubEnum(IntEnum):
    """Stub enum docstring."""

    STUB1 = auto()
    STUB2 = auto()


def many_edge_cases(
    x: int,
    y: None,
    union: int | None,
    pydantic_model: StubState,
    basic_dict: dict[str, int],
    complex_dict: dict[str, tuple[str, int]],
    enum: StubEnum,
    defaulted_str: str = "default",
    defaulted_float: float = 1.0,
) -> None:
    """
    Check using docstrings as partial f-string templates like so: {summary_format}.

    Args:
        x: Yes, I end with a colon :
        y: I am null.
            And despite that there is a multiline argument description.
        union: I am a union and the current year is {current_year}.
        pydantic_model: I am a Pydantic model.
        basic_dict: I am a dictionary with primitive values.
        complex_dict: I am a dictionary with complex values.
        enum: I am an enum.
        defaulted_str: I have a string default value.
        defaulted_float: I have a float default value.
    """


def add(a: int, b: int) -> int:
    """Add two numbers.

    Args:
        a: first number
        b: second number

    Returns:
        I am not yet included, perhaps someday I should be.
    """
    return a + b


def example_fxn(x: int, y: str, z: float) -> None:
    r"""A test function.

    There may be non-summary content.

    \f

    I should be ignored.

    Args:
        x: x
        y: y
        z: z
    """
    assert isinstance(x, int)
    assert isinstance(y, str)
    assert isinstance(z, float)


class TestTool:
    @pytest.mark.parametrize(
        ("fn", "kwargs", "expected"),
        [
            pytest.param(
                simple,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "simple",
                        "description": (
                            "Doing nothing may be better than doing something."
                        ),
                        "parameters": {
                            "properties": {},
                            "required": [],
                            "type": "object",
                        },
                    },
                },
                id="only-summary",
            ),
            pytest.param(
                intuitive_arg,
                {"allow_empty_param_descriptions": True},
                {
                    "type": "function",
                    "info": {
                        "name": "intuitive_arg",
                        "description": "Cast the input argument x to a float.",
                        "parameters": {
                            "properties": {"x": {"title": "X", "type": "string"}},
                            "required": ["x"],
                            "type": "object",
                        },
                    },
                },
                id="only-summary",
            ),
            pytest.param(
                many_edge_cases,
                {"current_year": 2024},  # Intentionally left format_1 unformatted,
                {
                    "type": "function",
                    "info": {
                        "name": "many_edge_cases",
                        "description": (
                            "Check using docstrings as partial f-string templates like"
                            " so: {summary_format}."
                        ),
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "x": {
                                    "description": "Yes, I end with a colon :",
                                    "title": "X",
                                    "type": "integer",
                                },
                                "y": {
                                    "description": (
                                        "I am null. And despite that there is a"
                                        " multiline argument description."
                                    ),
                                    "title": "Y",
                                    "type": "null",
                                },
                                "union": {
                                    "anyOf": [{"type": "integer"}, {"type": "null"}],
                                    "description": (
                                        "I am a union and the current year is 2024."
                                    ),
                                    "title": "Union",
                                },
                                "pydantic_model": {
                                    "$ref": "#/$defs/StubState",
                                    "description": "I am a Pydantic model.",
                                },
                                "basic_dict": {
                                    "additionalProperties": {"type": "integer"},
                                    "description": (
                                        "I am a dictionary with primitive values."
                                    ),
                                    "title": "Basic Dict",
                                    "type": "object",
                                },
                                "complex_dict": {
                                    "additionalProperties": {
                                        "maxItems": 2,
                                        "minItems": 2,
                                        "prefixItems": [
                                            {"type": "string"},
                                            {"type": "integer"},
                                        ],
                                        "type": "array",
                                    },
                                    "description": (
                                        "I am a dictionary with complex values."
                                    ),
                                    "title": "Complex Dict",
                                    "type": "object",
                                },
                                "enum": {
                                    "$ref": "#/$defs/StubEnum",
                                    "description": "I am an enum.",
                                },
                                "defaulted_str": {
                                    "default": "default",
                                    "description": "I have a string default value.",
                                    "title": "Defaulted Str",
                                    "type": "string",
                                },
                                "defaulted_float": {
                                    "default": 1.0,
                                    "description": "I have a float default value.",
                                    "title": "Defaulted Float",
                                    "type": "number",
                                },
                            },
                            "required": [
                                "x",
                                "y",
                                "union",
                                "pydantic_model",
                                "basic_dict",
                                "complex_dict",
                                "enum",
                            ],
                            "$defs": {
                                "StubEnum": {
                                    "description": "Stub enum docstring.",
                                    "enum": [1, 2],
                                    "title": "StubEnum",
                                    "type": "integer",
                                },
                                "StubState": {
                                    "description": "Stub model docstring.",
                                    "properties": {
                                        "defaulted_int": {
                                            "default": 1,
                                            "description": "A description of the int.",
                                            "title": "Defaulted Int",
                                            "type": "integer",
                                        },
                                        "required_str": {
                                            "description": "A description of the str.",
                                            "title": "Required Str",
                                            "type": "string",
                                        },
                                    },
                                    "required": ["required_str"],
                                    "title": "StubState",
                                    "type": "object",
                                },
                            },
                        },
                    },
                },
                id="many-edge-cases",
            ),
            pytest.param(
                add,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "add",
                        "description": "Add two numbers.",
                        "parameters": {
                            "properties": {
                                "a": {
                                    "description": "first number",
                                    "title": "A",
                                    "type": "integer",
                                },
                                "b": {
                                    "description": "second number",
                                    "title": "B",
                                    "type": "integer",
                                },
                            },
                            "required": ["a", "b"],
                            "type": "object",
                        },
                    },
                },
                id="with-args-and-returns",
            ),
            pytest.param(
                example_fxn,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "example_fxn",
                        "description": (
                            "A test function.\n\nThere may be non-summary content."
                        ),
                        "parameters": {
                            "properties": {
                                "x": {
                                    "description": "x",
                                    "title": "X",
                                    "type": "integer",
                                },
                                "y": {
                                    "description": "y",
                                    "title": "Y",
                                    "type": "string",
                                },
                                "z": {
                                    "description": "z",
                                    "title": "Z",
                                    "type": "number",
                                },
                            },
                            "required": ["x", "y", "z"],
                            "type": "object",
                        },
                    },
                },
                id="with-linefeed",
            ),
        ],
    )
    def test_from_function(
        self, fn: Callable, kwargs: dict[str, Any], expected: dict[str, Any]
    ) -> None:
        assert (
            Tool.from_function(fn, **kwargs).model_dump(exclude_none=True) == expected
        )

    @pytest.mark.parametrize(
        ("fn", "kwargs", "expected"),
        [
            (
                example_fxn,
                {},
                """NAME: example_fxn

SYNOPSIS:
    example_fxn(integer x, string y, number z)

DESCRIPTION:
    A test function.

    There may be non-summary content.

PARAMETERS:
    x (integer): x
    y (string): y
    z (number): z""",
            ),
            (
                intuitive_arg,
                {"allow_empty_param_descriptions": True},
                """NAME: intuitive_arg

SYNOPSIS:
    intuitive_arg(string x)

DESCRIPTION:
    Cast the input argument x to a float.

PARAMETERS:
    x (string): No description provided.""",
            ),
        ],
    )
    def test_describe_str(
        self, fn: Callable, kwargs: dict[str, Any], expected: str
    ) -> None:
        tool = Tool.from_function(fn, **kwargs)
        assert tool.info.describe_str().strip() == expected

    def test_describe(self, subtests: SubTests) -> None:
        """Test that describe_xyz functions for FunctionInfo are reasonable."""
        tool = Tool.from_function(example_fxn)

        with subtests.test("Test describe_xml is callable"):
            assert tool.info.describe_xml()

        with subtests.test("Test describe_json is callable"):
            assert tool.info.describe_json()

    def test_serialization_manual(self) -> None:
        # make one manually
        tool = Tool(
            tool_fn=add,
            info=FunctionInfo(
                name="add",
                description="Add two numbers.",
                parameters={
                    "properties": {
                        "a": {
                            "description": "first number",
                            "title": "A",
                            "type": "integer",
                        },
                        "b": {
                            "description": "second number",
                            "title": "B",
                            "type": "integer",
                        },
                    },
                    "required": ["a", "b"],
                    "type": "object",
                },
            ),
        )

        ref = json.loads(r"""{
        "type": "function",
        "function": {
            "name": "add",
            "description": "Add two numbers.",
            "parameters": {
                "type": "object",
                "properties": {
                    "a": {
                        "type": "integer",
                        "title": "A",
                        "description": "first number"
                    },
                    "b": {
                        "type": "integer",
                        "title": "B",
                        "description": "second number"
                    }
                },
                "required": [
                    "a",
                    "b"
                ]
            }
        }
    }""")
        # make sure it agrees with the reference
        my_dump = json.loads(tool.model_dump_json(exclude_none=True, by_alias=True))
        assert my_dump == ref

        # make one from a function
        tool_fxn = Tool.from_function(add)
        # make sure it serializes correctly
        assert tool_fxn.model_dump_json(
            exclude_none=True, by_alias=True
        ) == tool.model_dump_json(exclude_none=True, by_alias=True)

    @pytest.mark.asyncio
    async def test_arg_types(self) -> None:
        tool = Tool.from_function(example_fxn)

        assert tool.info.parameters.properties["x"]["type"] == "integer"
        assert tool.info.parameters.properties["y"]["type"] == "string"
        assert tool.info.parameters.properties["z"]["type"] == "number"

        calls = [
            ToolCall.from_name(tool.info.name, x=5, y="hi", z=4.2),
        ]
        for call in calls:
            # Call the function to make sure argument types
            # are passed correctly. Private because
            # it doesn't serialize
            tool._tool_fn(**call.function.arguments)

    @pytest.mark.asyncio
    async def test_tool_serialization(
        self, dummy_env: DummyEnv, subtests: SubTests
    ) -> None:
        def get_todo_list(n: int):
            """Get todo list for today.

            Args:
                n: number of items to return
            """
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"][:n])

        tool = Tool.from_function(get_todo_list)

        with subtests.test("pickling"):
            # Check round-trip pickling doesn't break the original Tool
            orig_tool_fn_id = id(tool._tool_fn)
            pickle.loads(pickle.dumps(tool))  # noqa: S301
            assert id(tool._tool_fn) == orig_tool_fn_id

        with subtests.test("serialization then deserialization"):
            tool_copy = Tool(**tool.model_dump(by_alias=True))
            assert tool.type == tool_copy.type
            assert tool.info == tool_copy.info

        dummy_env.tools = [tool]

        with subtests.test("tool call from dump"):
            # Mimic the way an ToolCall might be invoked by an LLM API:
            # the arguments will be strings.
            action = ToolRequestMessage(**{  # noqa: PIE804
                "tool_calls": [
                    {
                        "id": "good_tool_call",
                        "function": {"name": "get_todo_list", "arguments": '{"n": 2}'},
                    },
                    {
                        "id": "bad_tool_call",
                        "function": {
                            "name": "get_todo_list",
                            "arguments": '({"n": 2})',  # NOTE: invalid JSON
                        },
                    },
                ]
            })

            assert action.tool_calls[0].function.arguments == {"n": 2}
            assert action.tool_calls[1].function.name == INVALID_TOOL_NAME

        with subtests.test("tool call from name"):
            tool_call = ToolCall.from_name("get_todo_list", n=2)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk\nRead a book"

        with subtests.test("tool call from tool"):
            tool_call = ToolCall.from_tool(tool, n=2)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk\nRead a book"

        with subtests.test("tool call from tool with no kwargs"):
            tool_call = ToolCall.from_tool(tool, 3)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert (
                new_messages[0].content == "Go for a walk\nRead a book\nCall a friend"
            )

        def get_todo_list_no_args():
            """Get todo list for today."""
            return "Go for a walk"

        tool = Tool.from_function(get_todo_list_no_args)
        dummy_env.tools = [tool]

        with subtests.test("tool call from tool with no args and order mismatch"):
            tool_call = ToolCall.from_tool(tool)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk"

            tool_call = ToolCall.from_tool(tool, 1, 10, 30441)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk"


@pytest.mark.asyncio
async def test_argref_by_name_basic_usage() -> None:
    class MyState:
        def __init__(self):
            self.refs = {"foo": 1}

    # Check we can use argref_by_name to add 1 + 2 using a value in refs
    wrapped_add = argref_by_name(args_to_skip={"b"})(add)
    s = MyState()

    result = wrapped_add("foo", 2, state=s)
    # Now s.refs has a new entry at the below `name`
    name = result.split()[0]
    assert s.refs[name] == 1 + 2

    # Check kwargs work too
    result = wrapped_add(a="foo", b=2, state=s)
    name = result.split()[0]
    assert s.refs[name] == 1 + 2


@pytest.mark.asyncio
async def test_argref_by_name_error_handling() -> None:
    class MyState:
        def __init__(self):
            self.refs = {"foo": 1}

    wrapped_add = argref_by_name()(add)
    s = MyState()

    # Check if we use a key name that doesn't exist, we blow up
    with pytest.raises(KeyError, match="not found in state"):
        wrapped_add("bar", 2, state=s)

    # Check if state doesn't have refs, we blow up
    with pytest.raises(AttributeError, match="must have a 'refs' attribute"):
        wrapped_add("foo", 2, state="not a state")

    # Check that we cannot pass a direct value as a kwarg
    with pytest.raises(KeyError, match="Key is not present"):
        wrapped_add(a=1, b=2, state=s)


@pytest.mark.asyncio
async def test_argref_by_name_async_functions() -> None:
    class MyState:
        def __init__(self):
            self.refs = {"foo": 1, "bar": 7}

    # Define the async_add function with the decorator
    @argref_by_name()
    async def async_add(a: int, b: int) -> int:
        """Some docstring."""
        return a + b

    s = MyState()
    result = await async_add("foo", 2, state=s)
    assert s.refs[result.split()[0]] == 1 + 2

    result = await async_add(6, 2, state=s)
    assert s.refs[result.split()[0]] == 6 + 2

    # Now try with lists
    result = await async_add("foo", "bar", state=s)
    assert s.refs[result.split()[0]] == 1 + 7

    # Try the convenience of comma splitting on key
    result = await async_add("foo,bar", state=s)
    assert s.refs[result.split()[0]] == 1 + 7

    # Define and test async_list
    @argref_by_name()
    async def async_list(a: int, b: int) -> list[int]:
        """Some docstring."""
        return [a, b]

    result = await async_list("foo", 2, state=s)
    name1, name2 = (n.split()[0] for n in result.split("\n"))
    assert s.refs[name1] == 1
    assert s.refs[name2] == 2

    # Define and test async_list_direct
    @argref_by_name(return_direct=True)
    async def async_list_direct(a: int, b: int) -> list[int]:
        """Some docstring."""
        return [a, b]

    assert await async_list_direct("foo", 2, state=s) == [1, 2]


@pytest.mark.asyncio
async def test_argref_by_name_advanced_features() -> None:
    class MyState:
        def __init__(self):
            self.refs = {"foo": 1}

    s = MyState()

    # Define and test dereference via no state value found with return_direct
    @argref_by_name(return_direct=True)
    def skip_deref_test(foo: float, a: str) -> str:
        """Some docstring."""
        return f"{foo} {a}"

    assert skip_deref_test("foo", "not in state", state=s) == "1 not in state"
    assert skip_deref_test("foo", "foo", state=s) == "1 1"

    # Call in context using Tool and related classes
    wrapped_add = argref_by_name(args_to_skip={"b"})(add)
    tool = Tool.from_function(wrapped_add)

    tool_call = ToolCall.from_tool(tool, "foo", b=2)
    action = ToolRequestMessage(tool_calls=[tool_call])
    my_env = DummyEnv()
    my_env.tools = [tool]
    new_messages = await my_env.exec_tool_calls(action, state=MyState())
    assert new_messages[0].content.endswith("3")

    # Assert that we can describe the tool
    assert tool.info.describe_str()
    assert "(Pass a string key instead of the full object)" in tool.info.describe_str()

    # Test state passing with fxn_requires_state
    @argref_by_name(fxn_requires_state=True)
    async def want_state(a: int, state: MyState) -> int:  # noqa: ARG001
        """Some docstring.

        Args:
            a: first number
            state: the state object
        """
        return 2 * a

    tool = Tool.from_function(want_state)
    action = ToolRequestMessage(tool_calls=[ToolCall.from_tool(tool, "foo")])
    my_env = DummyEnv()
    my_env.tools = [tool]
    await my_env.exec_tool_calls(action, state=MyState())

    # Check we can pass kwarg lists as comma-separated keys
    @argref_by_name(return_direct=True)
    def kwarg_list_test(a: list[int]) -> int:
        return sum(a)

    assert kwarg_list_test(a="foo,foo", state=s) == 2


@pytest.mark.asyncio
async def test_argref_by_name_type_checking() -> None:
    class MyInt(int):
        pass

    class MyState:
        def __init__(self):
            self.refs = {
                "int_arg": 1,
                "str_arg": "abc",
                "int_list_arg": [1],
                "str_list_arg": ["abc"],
                "my_int_list_arg": [MyInt()],
            }

    s = MyState()

    def typed_fn(a: int, b) -> int:  # noqa: ARG001
        """Some docstring."""
        return a

    # Make sure we can decorate the function twice. Decoration should not
    # modify the underlying function or its annotations.
    for _ in range(2):
        type_checked_fn = argref_by_name(type_check=True)(typed_fn)

        type_checked_fn(a="int_arg", b="str_arg", state=s)  # correctly-typed
        with pytest.raises(TypeError):
            # A non-int value is passed to a by name
            type_checked_fn(a="str_arg", b="str_arg", state=s)

    def complex_typed_fn(c: Sequence[int], d: int | str) -> None:
        """Some docstring."""

    for _ in range(2):
        type_checked_fn = argref_by_name(type_check=True)(complex_typed_fn)

        type_checked_fn(c="int_list_arg", d="str_arg", state=s)  # correctly-typed
        # list[MyInt] should match Sequence[int]
        type_checked_fn(c="my_int_list_arg", d="str_arg", state=s)

        with pytest.raises(TypeError):
            # passing int, not list[int]
            type_checked_fn(c="int_arg", d="str_arg", state=s)

        with pytest.raises(TypeError):
            # passing list[str], not list[int]
            type_checked_fn(c="str_list_arg", d="int_arg", state=s)


@pytest.mark.asyncio
async def test_make_tool_server():
    def add(a: int, b: int) -> int:
        """Add two numbers."""
        return a + b

    def subtract(a: int, b: int) -> int:
        """Subtract two numbers.

        Args:
            a: first number
            b: second number
        """
        return a - b

    class MyEnv(Environment):
        async def reset(self) -> tuple[list[Message], list[Tool]]:
            tools = [
                Tool.from_function(add, allow_empty_param_descriptions=True),
                Tool.from_function(subtract),
            ]
            self.tools = tools
            return [], tools

        async def step(self, action):
            return await self.exec_tool_calls(action), False, 0, 0

        async def export_frame(self):
            pass

    with suppress_type_checks():
        server = await make_tool_server(MyEnv)

    # make sure there are two endpoints
    route_names = [route.name for route in server.routes]
    assert "add" in route_names
    assert "subtract" in route_names

    # make sure we can call them
    client = TestClient(server)
    token = "stub"
    with patch.dict(os.environ, {"AUTH_TOKEN": token}):
        response = client.post(
            "/add", json={"a": 1, "b": 2}, headers={"Authorization": f"Bearer {token}"}
        )
        assert response.status_code == 200
        assert response.json()["result"] == "3"



import logging
from abc import ABC, abstractmethod
from typing import Any, Generic, Self, TypeVar, cast

import httpx
from pydantic import BaseModel, Field

from aviary.env import Environment, TaskDataset
from aviary.message import Message
from aviary.tools import MessagesAdapter, Tool, ToolRequestMessage, ToolsAdapter

logger = logging.getLogger(__name__)

# Not sure why, but mypy complains if we use the TEnvState in aviary.env, so redefine here
TEnvState = TypeVar("TEnvState")


class EnvironmentClient(Environment[TEnvState], ABC, Generic[TEnvState]):
    def __init__(
        self,
        reset_endpoint_url: str,
        step_endpoint_url: str,
        request_params: httpx._types.QueryParamTypes | None = None,
        request_headers: httpx._types.HeaderTypes | None = None,
        request_timeout: float | None = None,
    ):
        self._reset_request_url = reset_endpoint_url
        self._step_request_url = step_endpoint_url
        self._request_params = request_params
        self._request_headers = request_headers
        self._request_timeout = request_timeout

    async def _post(self, url: str, json: dict[str, Any]) -> httpx.Response:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                json=json,
                params=self._request_params,
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return response

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        response = await self._post(
            self._reset_request_url, json=self._make_post_json(self.state)
        )
        msgs, tools = response.json()
        return (
            MessagesAdapter.validate_python(msgs),
            ToolsAdapter.validate_python(tools),
        )

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[list[Message], float, bool, bool]:
        response = await self._post(
            self._step_request_url,
            json=self._make_post_json(self.state) | {"action": action.model_dump()},
        )
        messages, reward, done, truncated = response.json()
        return MessagesAdapter.validate_python(messages), reward, done, truncated

    @abstractmethod
    def _make_post_json(self, state: TEnvState) -> dict[str, Any]:
        """Extract values from state to sent as JSON for all reset/step POSTs."""


class TaskEnvClientState(BaseModel):
    env_id: str = Field(
        description="The ID of the environment (provided by server on start)."
    )


class TaskEnvironmentClient(EnvironmentClient[TaskEnvClientState]):
    """An environment client for environments created by a TaskDatasetServer."""

    def __init__(self, idx: int | None, base_url: str, **kwargs):
        self._idx = idx
        self._start_request_url = base_url + "/start"
        self._close_request_url = base_url + "/close"

        kwargs = {
            "reset_endpoint_url": base_url + "/reset",
            "step_endpoint_url": base_url + "/step",
        } | kwargs

        super().__init__(**kwargs)

    async def _start_remote_env(self) -> str:
        response = await self._post(
            self._start_request_url, json={"task_idx": self._idx}
        )
        return response.json()["env_id"]

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        # defer starting to reset so we can make it async and set the state
        env_id = await self._start_remote_env()
        self.state = TaskEnvClientState(env_id=env_id)

        return await super().reset()

    async def close(self) -> None:
        if not hasattr(self, "state"):
            logger.warning("Attempting to close an environment that was never started.")
            return None

        response = await self._post(
            self._close_request_url, json=self._make_post_json(self.state)
        )
        return response.json()  # noqa: FURB184

    def _make_post_json(self, state: TaskEnvClientState) -> dict[str, Any]:
        return {"env_id": state.env_id}


UNSET_LEN = object()  # sentinel


class TaskDatasetClient(TaskDataset[TaskEnvironmentClient]):
    def __init__(
        self,
        server_url: str,
        # Note that None means no timeout, which is not a good default
        request_timeout: float | None = 300.0,
    ):
        self.server_url = server_url
        self.request_timeout = request_timeout
        self._len: int | object | None = UNSET_LEN

    def get_http_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(base_url=self.server_url, timeout=self.request_timeout)

    @classmethod
    async def create(cls, *args, **kwargs) -> Self:
        # We need to make an async request to the server to get the dataset size,
        # so provide this classmethod to instantiate the client & get the size.
        client = cls(*args, **kwargs)
        async with client.get_http_client() as http_client:
            client._len = (await http_client.get("/info")).json()["dataset_size"]
        return client

    def get_new_env_by_idx(self, idx: int) -> TaskEnvironmentClient:
        return self._make_env_client(idx)

    def get_new_env(self) -> TaskEnvironmentClient:
        return self._make_env_client(None)

    def _make_env_client(self, idx: int | None) -> TaskEnvironmentClient:
        return TaskEnvironmentClient(
            idx=idx, base_url=self.server_url, request_timeout=self.request_timeout
        )

    def __len__(self) -> int:
        if self._len is UNSET_LEN:
            raise RuntimeError(
                "Dataset should be created using `TaskDatasetClient.create()`"
            )
        if self._len is None:
            raise TypeError("Server did not define dataset length.")

        return cast(int, self._len)



import pathlib
from enum import StrEnum


class CILLMModelNames(StrEnum):
    """Models to use for generic CI testing."""

    ANTHROPIC = "claude-3-haiku-20240307"  # Cheap and not Anthropic's cutting edge
    OPENAI = "gpt-4o-mini-2024-07-18"  # Cheap and not OpenAI's cutting edge


TESTS_DIR = pathlib.Path(__file__).parent
CASSETTES_DIR = TESTS_DIR / "cassettes"



import uuid
from collections.abc import Callable, Mapping, Sequence
from functools import update_wrapper
from inspect import signature
from itertools import starmap
from types import UnionType
from typing import Any, Union, get_args, get_origin

from docstring_parser import compose, parse

from aviary.utils import is_coroutine_callable


def make_pretty_id(prefix: str = "") -> str:
    """
    Get an ID that is made using an optional prefix followed by part of an uuid4.

    For example:
    - No prefix: "ff726cd1"
    - With prefix of "foo": "foo-ff726cd1"
    """
    uuid_frags: list[str] = str(uuid.uuid4()).split("-")
    if not prefix:
        return uuid_frags[0]
    return prefix + "-" + uuid_frags[0]


DEFAULT_ARGREF_NOTE = "(Pass a string key instead of the full object)"
LIST_ARGREF_NOTE = "(Pass comma-separated string keys instead of the full object)"


def argref_wrapper(wrapper, wrapped, args_to_skip: set[str]):
    """Inject the ARGREF_NOTE into the Args."""
    # normal wraps
    wrapped_func = update_wrapper(wrapper, wrapped)
    # when we modify wrapped_func's annotations, we don't want to mutate wrapped
    wrapped_func.__annotations__ = wrapped_func.__annotations__.copy()
    # now adjust what we need
    for a in wrapped_func.__annotations__:
        if a in args_to_skip:
            continue
        wrapped_func.__annotations__[a] = str

    orig_annots = wrapped.__annotations__

    # now add note to docstring for all relevant Args
    if wrapped_func.__doc__:
        ds = parse(wrapped_func.__doc__)
        for param in ds.params:
            if param.arg_name in args_to_skip:
                continue

            note = DEFAULT_ARGREF_NOTE

            if (
                param.type_name is None
                and (type_hint := orig_annots.get(param.arg_name)) is not None
            ):
                param.type_name = _type_to_str(type_hint)
                if list in {type_hint, get_origin(type_hint)}:
                    note = LIST_ARGREF_NOTE

            param.description = (param.description or "") + f" {note}"

        wrapped_func.__doc__ = compose(ds)

    return wrapped_func


def argref_by_name(  # noqa: PLR0915
    fxn_requires_state: bool = False,
    prefix: str = "",
    return_direct: bool = False,
    type_check: bool = False,
    args_to_skip: set[str] | None = None,
):
    """Decorator to allow args to be a string key into a refs dict instead of the full object.

    This can prevent LLM-powered tool selections from getting confused by full objects,
    instead it enables them to work using named references. If a reference is not found, it
    will fallback on passing the original argument unless it is the first argument. If the
    first argument str is not found in the state object, it will raise an error.

    Args:
        fxn_requires_state: Whether to pass the state object to the decorated function.
        prefix: A prefix to add to the generated reference ID.
        return_direct: Whether to return the result directly or update the state object.
        type_check: Whether to type-check arguments with respect to the wrapped function's
            type annotations.
        args_to_skip: If provided, a set of argument names that should not be referenced by name.

    Example 1:
        >>> @argref_by_name()  # doctest: +SKIP
        >>> def my_func(foo: float): ...  # doctest: +SKIP

    Example 2:
        >>> def my_func(foo: float, bar: float) -> list[float]:
        ...     return [foo, bar]
        >>> wrapped_fxn = argref_by_name()(my_func)
        >>> # Equivalent to my_func(state.refs["foo"])
        >>> wrapped_fxn("foo", state=state)  # doctest: +SKIP

    Working with lists:
    - If you return a list, the decorator will create a new reference for each item in the list.
    - If you pass multiple args that are strings, the decorator will assume those are the keys.
    - If you need to pass a string, then use a keyword argument.

    Example 1:
        >>> @argref_by_name()  # doctest: +SKIP
        >>> def my_func(foo: float, bar: float) -> list[float]:  # doctest: +SKIP
        ...     return [foo, bar]  # doctest: +SKIP

    Example 2:
        >>> def my_func(foo: float, bar: float) -> list[float]:
        ...     return [foo, bar]
        >>> wrapped_fxn = argref_by_name()(my_func)
        >>> # Returns a multiline string with the new references
        >>> # Equivalent to my_func(state.refs["a"], state.refs["b"])
        >>> wrapped_fxn("a", "b", state=state)  # doctest: +SKIP
    """
    args_to_skip = (args_to_skip or set()) | {"state", "return"}

    def decorator(func):  # noqa: PLR0915
        def get_call_args(*args, **kwargs):
            if "state" not in kwargs:
                raise ValueError(
                    "argref_by_name decorated function must have a 'state' argument."
                    " Function signature:"
                    f" {func.__name__}({', '.join(func.__annotations__)})  received"
                    f" args: {args} kwargs: {kwargs}"
                )
            # pop the state argument
            state = kwargs["state"] if fxn_requires_state else kwargs.pop("state")

            # now convert the keynames to actual references (if they are a string)
            # tuple is (arg, if was dereferenced)
            def maybe_deref_arg(arg, must_exist: bool) -> tuple[Any, bool]:
                try:
                    refs = state.refs
                except AttributeError as e:
                    raise AttributeError(
                        "The state object must have a 'refs' attribute to use"
                        " argref_by_name decorator."
                    ) from e

                if arg in refs:
                    return [refs[arg]], True

                # sometimes it is not correctly converted to a tuple
                # so as an attempt to be helpful...
                if (
                    isinstance(arg, str)
                    and len(split_args := [a.strip() for a in arg.split(",")]) > 1
                ):
                    if not (missing := [a for a in split_args if a not in refs]):
                        return [refs[a] for a in split_args], True

                    if must_exist:
                        # Error message for the agent - cast back to comma-separated, since that's the format the agent
                        # is expected to use.
                        raise KeyError(
                            "The following keys are not present in the current"
                            f' key-value store: "{", ".join(missing)}"'
                        )

                if not must_exist:
                    return arg, False

                # Error message for the agent
                raise KeyError(
                    f"Key is not present in the current key-value store: {arg!r}"
                )

            # the split thing makes it complicated and we cannot use comprehension
            deref_args = []
            for i, arg in enumerate(args):
                # In order to support *args, allow arguments that are either ref keys or strings
                a, dr = maybe_deref_arg(arg, must_exist=False)
                if dr:
                    deref_args.extend(a)
                else:
                    if i == 0 and isinstance(arg, str):
                        # This is a bit of a heuristic, but if the first arg is a string and not found
                        # likely the user intended to use a reference
                        raise KeyError(f"The key {arg} is not found in state.")
                    deref_args.append(a)

            deref_kwargs = {}
            for k, v in kwargs.items():
                if args_to_skip and k in args_to_skip:
                    deref_kwargs[k] = v
                    continue

                # In the kwarg case, force arguments to be ref keys (unless in args_to_skip)
                a, _ = maybe_deref_arg(v, must_exist=True)
                if len(a) > 1:
                    # We got multiple items, so pass the whole list
                    deref_kwargs[k] = a
                else:
                    # We only got one item - pass it directly
                    deref_kwargs[k] = a[0]

            return deref_args, deref_kwargs, state

        def update_state(state, result):
            if return_direct:
                return result
            # if it returns a list, rather than storing the list as a single reference
            # we store each item in the list as a separate reference
            if isinstance(result, list):
                msg = []
                for item in result:
                    new_name = make_pretty_id(prefix)
                    state.refs[new_name] = item
                    msg.append(f"{new_name} ({item.__class__.__name__}): {item!s}")
                return "\n".join(msg)
            new_name = make_pretty_id(prefix)
            state.refs[new_name] = result
            return f"{new_name} ({result.__class__.__name__}): {result!s}"

        def wrapper(*args, **kwargs):
            args, kwargs, state = get_call_args(*args, **kwargs)
            if type_check:
                _check_arg_types(func, args, kwargs)
            result = func(*args, **kwargs)
            return update_state(state, result)

        async def awrapper(*args, **kwargs):
            args, kwargs, state = get_call_args(*args, **kwargs)
            if type_check:
                _check_arg_types(func, args, kwargs)
            result = await func(*args, **kwargs)
            return update_state(state, result)

        if is_coroutine_callable(func):
            awrapper = argref_wrapper(awrapper, func, args_to_skip)
            awrapper.requires_state = True  # type: ignore[attr-defined]
            return awrapper

        wrapper = argref_wrapper(wrapper, func, args_to_skip)
        wrapper.requires_state = True  # type: ignore[attr-defined]
        return wrapper

    return decorator


def _check_arg_types(func: Callable, args, kwargs) -> None:
    annotations = {
        k: v for k, v in func.__annotations__.items() if k not in {"return", "state"}
    }

    sig = signature(func)
    param_names = list(sig.parameters.keys())

    # Elements are tuple of (param, expected, provided)
    wrong_types: list[tuple[str, str, str]] = []

    # Map positional arguments to their parameter names
    for idx, arg in enumerate(args):
        if idx >= len(param_names):
            break  # Extra arguments are handled by *args if any
        param = param_names[idx]
        expected_type = annotations.get(param)
        if expected_type and not _isinstance_with_generics(arg, expected_type):
            wrong_types.append((
                param,
                _type_to_str(expected_type),
                _type_to_str(type(arg)),
            ))

    # Check keyword arguments
    for param, arg in kwargs.items():
        expected_type = annotations.get(param)
        if expected_type and not _isinstance_with_generics(arg, expected_type):
            wrong_types.append((
                param,
                # sometimes need str for generics like Union
                _type_to_str(expected_type),
                _type_to_str(type(arg)),
            ))

    if wrong_types:
        raise TypeError(
            "The following arguments have incorrect types:\n"
            + "\n".join(
                f"- {param}: expected {expected}, got {provided}"
                for param, expected, provided in wrong_types
            )
        )


def _type_to_str(t) -> str:
    """
    Convert a Python type annotation into its string representation.

    Examples:
        type_to_str(int) -> "int"
        type_to_str(Union[int, float]) -> "int | float"
        type_to_str(list[str]) -> "list[str]"
    """
    origin = get_origin(t)
    args = get_args(t)

    if origin is Union:
        # Handle Union types, including the new | syntax in Python 3.10+
        return " | ".join(_type_to_str(arg) for arg in args)
    if origin is not None:
        # Handle generic types like list[str], dict[str, int], etc.
        origin_name = origin.__name__
        args_str = ", ".join(_type_to_str(arg) for arg in args)
        return f"{origin_name}[{args_str}]"
    if hasattr(t, "__name__"):
        # Handle basic types
        return t.__name__
    # Fallback for types without a __name__ attribute
    return str(t)


def _isinstance_with_generics(obj, expected_type) -> bool:  # noqa: PLR0911
    """Like isinstance, but with support for generics."""
    origin = get_origin(expected_type)
    if origin is None:
        # Handle special cases like typing.Any
        if expected_type is Any:
            return True
        return isinstance(obj, expected_type)
    if origin in {UnionType, Union}:
        return any(
            _isinstance_with_generics(obj, arg) for arg in get_args(expected_type)
        )
    if origin in {list, Sequence}:
        if not isinstance(obj, Sequence):
            return False
        elem_type = get_args(expected_type)[0]
        return all(_isinstance_with_generics(elem, elem_type) for elem in obj)
    if origin in {dict, Mapping}:
        if not isinstance(obj, Mapping):
            return False
        key_type, val_type = get_args(expected_type)
        return all(
            _isinstance_with_generics(k, key_type)
            and _isinstance_with_generics(v, val_type)
            for k, v in obj.items()
        )
    if origin is tuple:
        if not isinstance(obj, tuple):
            return False
        elem_types = get_args(expected_type)
        if len(elem_types) == 2 and elem_types[1] is Ellipsis:  # noqa: PLR2004
            # Tuple of variable length
            return all(_isinstance_with_generics(elem, elem_types[0]) for elem in obj)
        if len(elem_types) != len(obj):
            return False
        return all(
            starmap(_isinstance_with_generics, zip(obj, elem_types, strict=True))
        )
    # Fallback to checking the origin type
    return isinstance(obj, origin)



from typing import Any

import pytest

from aviary.core import DummyEnv

from . import CASSETTES_DIR


@pytest.fixture(name="dummy_env")
def fixture_dummy_env() -> DummyEnv:
    return DummyEnv()


OPENAI_API_KEY_HEADER = "authorization"
ANTHROPIC_API_KEY_HEADER = "x-api-key"
# SEE: https://github.com/kevin1024/vcrpy/blob/v6.0.1/vcr/config.py#L43
VCR_DEFAULT_MATCH_ON = "method", "scheme", "host", "port", "path", "query"


@pytest.fixture(scope="session", name="vcr_config")
def fixture_vcr_config() -> dict[str, Any]:
    return {
        "filter_headers": [OPENAI_API_KEY_HEADER, ANTHROPIC_API_KEY_HEADER, "cookie"],
        "record_mode": "once",
        "match_on": ["method", "host", "path", "query"],
        "allow_playback_repeats": True,
        "cassette_library_dir": str(CASSETTES_DIR),
    }



import base64
import contextlib
import inspect
import io
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np


def partial_format(value: str, **formats: dict[str, Any]) -> str:
    """Partially format a string given a variable amount of formats."""
    for template_key, template_value in formats.items():
        with contextlib.suppress(KeyError):
            value = value.format(**{template_key: template_value})
    return value


def encode_image_to_base64(img: "np.ndarray") -> str:
    """Encode an image to a base64 string, to be included as an image_url in a Message."""
    try:
        from PIL import Image
    except ImportError as e:
        raise ImportError(
            "Image processing requires the 'image' extra for 'Pillow'. Please:"
            " `pip install aviary[image]`."
        ) from e

    image = Image.fromarray(img)
    buffer = io.BytesIO()
    image.save(buffer, format="PNG")
    return (
        f"data:image/png;base64,{base64.b64encode(buffer.getvalue()).decode('utf-8')}"
    )


def is_coroutine_callable(obj) -> bool:
    """Get if the input object is awaitable."""
    if inspect.isfunction(obj) or inspect.ismethod(obj):
        return inspect.iscoroutinefunction(obj)
    if callable(obj):
        return inspect.iscoroutinefunction(obj.__call__)
    return False



"""HotPotQA environment for aviary agents.

Implements the HotPotQA multihop question-answering environment from:

Yao et al., ReAct: Synergizing Reasoning and Acting in Language Models.
In The Eleventh International Conference on Learning Representations. 2023

Agents in the HotPotQA environment can perform searches and lookups to retrieve specific information.
The environment supports 3 tools:

search[entity]: Search for a specific entity on Wikipedia and return relevant information.
lookup[keyword]: Find and return the next sentence containing the keyword in the current passage.
submit_answer[answer]: Submit the final answer to the question and conclude the task.
"""

import logging
import os
import random
import re
import string
from collections.abc import Callable
from enum import StrEnum
from typing import Any, ClassVar, cast

import httpx
from bs4 import BeautifulSoup
from datasets import load_dataset
from pydantic import BaseModel, ConfigDict, Field, field_validator
from tenacity import retry, stop_after_attempt, wait_exponential_jitter

from aviary.core import (
    Environment,
    EvalAnswerMode,
    Frame,
    Message,
    Messages,
    TaskDataset,
    Tool,
    ToolRequestMessage,
    eval_answer,
)

logger = logging.getLogger(__name__)


# Jitter in case we have lots of requests at once, to space them out slightly
@retry(stop=stop_after_attempt(2), wait=wait_exponential_jitter(initial=1, max=4))
async def fetch_with_retry(
    client: httpx.AsyncClient, url: str, **kwargs
) -> httpx.Response:
    response = await client.get(url, **kwargs)
    if response.status_code == httpx.codes.TOO_MANY_REQUESTS:
        response.raise_for_status()
    return response


class HotPotQAEnvState(BaseModel):
    """State of the HotPotQA environment."""

    done: bool = Field(
        default=False,
        description="Flag for being done, automatically set true after many steps.",
    )

    steps: int = Field(default=0, description="Count of environment steps.")

    reward: float = Field(
        default=0.0, description="Current reward value, reset each environment step."
    )
    answer: str | None = Field(
        default=None,
        description="The answer to the question, or None if not yet answered.",
    )
    last_action_is_lookup: bool = Field(
        default=False,
        description="Whether the last action was a lookup action."
        "Default is False, as after reset the agent has not yet taken any action.",
    )
    last_lookup: str | None = Field(
        default=None, description="The last lookup keyword."
    )
    lookup_results: list[str] = Field(
        default_factory=list, description="Results of the last lookup."
    )
    lookup_index: int = Field(
        default=0, description="Index of the last retrieved lookup result."
    )
    page: str | None = Field(default=None, description="The current Wikipedia page.")


def create_tool(function: Callable, name: str) -> Tool:
    """Create a Tool object from a function and set its name.

    Args:
        function: The function to be wrapped by the Tool.
        name: The name to assign to the Tool.

    Returns:
        A Tool object with the specified function and name.
    """
    tool = Tool.from_function(function)
    tool.info.name = name
    return tool


def clean_str(p: str) -> str:
    r"""Clean and normalize a given string by encoding and decoding it.

    Args:
        p: The input string to be cleaned and normalized.

    Returns:
        The cleaned and normalized UTF-8 encoded string.

    Examples:
        >>> clean_str("This is a test string with unicode escape: \\u00e9")
        'This is a test string with unicode escape: é'
        >>> clean_str(r"SBN IT\ICCU\UBO\2771748.")  # SEE: https://en.wikipedia.org/wiki/Tricolour_Day
        'SBN IT\\ICCU\\UBO\\2771748.'
    """
    # Original source:
    # https://github.com/ysymyth/ReAct/blob/6bdb3a1fd38b8188fc7ba4102969fe483df8fdc9/wikienv.py#L10-L11
    try:
        # unicode-escape codec can interpret backslash escapes (e.g., `\n`, `\u1234`)
        return (
            p.encode("latin1").decode("unicode-escape").encode("utf-8").decode("utf-8")
        )
    except (UnicodeEncodeError, UnicodeDecodeError):
        logger.debug(f"Replacing bad characters in input string {p!r}.")
        return p.encode(errors="replace").decode("utf-8", errors="replace")


def normalize_answer(s: str | float) -> str:
    """Normalize the given answer string by applying several text processing steps.

    This function processes the input string through a series of transformations to
    normalize it, which includes:
        1. Convert all characters to lowercase.
        2. Remove punctuation from the text.
        3. Remove articles ('a', 'an', 'the') from the text.
        4. Fix extra whitespace by reducing multiple spaces to a single space.

    Args:
        s: The input answer string to be normalized.

    Returns:
        The normalized answer string.

    Example:
        >>> normalize_answer("The Quick, Brown Fox!")
        'quick brown fox'
    """
    # if s is not a string, convert it to a string
    if not isinstance(s, str):
        return str(s)

    def remove_articles(text: str) -> str:
        """Remove articles ('a', 'an', 'the') from the text."""
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text: str) -> str:
        """Reduce multiple spaces to a single space."""
        return " ".join(text.split())

    def remove_punc(text: str) -> str:
        """Remove all punctuation from the text."""
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text: str) -> str:
        """Convert all characters to lowercase."""
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


class HotPotQAEnv(Environment[HotPotQAEnvState]):
    State: ClassVar = HotPotQAEnvState
    wiki_cache: ClassVar[dict[str, str]] = {}

    def __init__(
        self,
        question: str,
        correct_answer: str | float,
        correct_reward: float = 1.0,
        incorrect_reward: float = 0.0,
        tool_failure_reward: float = 0.0,
        evaluation_mode: EvalAnswerMode = EvalAnswerMode.CONTAINS,
        proxy: str | None = None,
    ):
        super().__init__()
        self.question = question
        # Normalize the correct answer once here as a minor performance optimization
        self.normalized_correct_answer = normalize_answer(correct_answer)
        self.correct_reward = correct_reward
        self.incorrect_reward = incorrect_reward
        self.tool_failure_reward = tool_failure_reward
        self.proxy = proxy
        self.evaluation_mode = evaluation_mode

        if evaluation_mode == EvalAnswerMode.LLM_SCORE:
            raise NotImplementedError(
                f'{HotPotQAEnv.__name__} does not support "{evaluation_mode}"'
                " since the environment was built around binary evaluation of the"
                " answer. Further development is needed for this mode."
            )

        # Title case tool names to match third party demonstration data
        self.tools = [
            Tool.from_function(self.search),
            Tool.from_function(self.lookup),
            Tool.from_function(self.submit_answer),
        ]

    @classmethod
    def from_task(cls, task: str) -> "HotPotQAEnv":
        return cls(question=task, correct_answer=0.0)

    async def calculate_answer_reward(self, answer: str | None) -> float:
        """Calculate the reward based on the agent's answer.

        Returns:
            The correct reward if the agent's answer is correct (prediction exactly
                matches ground truth), otherwise the incorrect reward.
        """
        if answer is None:
            return self.incorrect_reward
        return (
            self.correct_reward
            if (
                await eval_answer(
                    normalize_answer(answer),
                    self.normalized_correct_answer,
                    self.evaluation_mode,
                )
            )
            else self.incorrect_reward
        )

    async def reset(self) -> tuple[Messages, list[Tool]]:
        """Reset the HotPotQA environment to an initial state.

        This method resets the environment to its initial state, setting up necessary variables and tools
        for the agent to interact with. It prepares the environment for a new episode by initializing
        various attributes and returning the initial observation and tools available for the agent.

        Returns:
            tuple: A tuple containing:
                - Messages: The initial observation wrapped in a Message object.
                - list[Tool]: A list of tools (search, lookup, and submit_answer) available for the agent.

        Example:
            >>> env = HotPotQAEnv()
            >>> initial_obs, tools = env.reset(seed=42, idx=5)
            >>> print(initial_obs)
            [Message(content='Question: <question_text>')]
            >>> print(tools)
            [<Tool: search>, <Tool: lookup>, <Tool: submit_answer>]
        """
        self.state = self.State()
        return [Message(content=f"Question: {self.question}")], self.tools

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[Messages, float, bool, bool]:
        """Take a step in the environment. Assume only one tool at a time can be called for HotpotQA.

        This method processes an action message, which can be a tool request, a finish request, or an error message.
        It updates the environment state accordingly and returns the observation and done status.

        Args:
            action: Action to take.

        Returns:
            Tuple[Messages, bool]: A tuple containing:
                - Messages: The response message(s) from the executed tool.
                - bool: The done status indicating whether the episode is finished.

        Example:
            >>> env = HotPotQAEnv()
            >>> action = ToolRequestMessage(
            ...     tool_calls=[
            ...         ToolCall(
            ...             function=ToolCallFunction(name="Search"),
            ...             arguments={"entity": "Python"},
            ...         )
            ...     ]
            ... )
            >>> obs, done = await env.step(action)
            >>> print(obs, done)
            [ToolResponseMessage(name='Search', tool_call_id='tool_call_id', content='...')], False
        """
        self.state.steps += 1
        if not action.tool_calls:
            return (
                [
                    Message(
                        content=(
                            "Must call one of the provided tools"
                            f" { {t.info.name for t in self.tools} }."
                        )
                    )
                ],
                self.tool_failure_reward,
                self.state.done,
                False,
            )

        # We accumulate reward across all tool calls in a step
        self.state.reward = 0.0
        response_messages = cast(
            Messages,
            # Ordered since things like search -> lookup need to be run in order.
            # NOTE: Handling tool exceptions here keeps the trajectory going, but I don't
            # think the returned message is useful to the agent/learning. Disabling for now.
            await self.exec_tool_calls(
                action,
                ordered=True,
            ),
        )
        return response_messages, self.state.reward, self.state.done, False

    def export_frame(self) -> Frame:
        """Export the current state of the environment as a Frame object.

        This method creates and returns a Frame object that captures the current state and additional
        information of the HotPotQA environment. This can be useful for logging, debugging, or visualization
        purposes.

        Returns:
            Frame: A Frame object containing the current state and additional information of the environment.

        The returned Frame object includes the following information:
            - state (dict): An empty dictionary representing the current state of the environment.
            - info (dict): A dictionary containing:
                - "steps" (int): The number of steps taken in the current episode.
                - "done" (bool): A flag indicating whether the episode is finished.
                - "reward" (float): The accumulated reward in the current episode.
                - "answer" (Optional[str]): The answer provided by the agent, if any.

        Example:
            >>> env = HotPotQAEnv()
            >>> frame = env.export_frame()
            >>> print(frame.info)
            {'steps': 0, 'done': False, 'reward': 0.0, 'answer': None}
        """
        return Frame(
            info={
                "steps": self.state.steps,
                "done": self.state.done,
                "reward": self.state.reward,
                "answer": self.state.answer,
            }
        )

    async def submit_answer(self, answer: str) -> str:
        """Finish the task by submitting an answer to the question.

        Args:
            answer: The answer to the question.
        """
        self.state.done = True
        if not answer:
            return "submit_answer failed since it was provided an empty answer."

        self.state.answer = answer
        self.state.reward += await self.calculate_answer_reward(answer)

        self.state.last_action_is_lookup = False
        return "Finished."

    async def search(self, entity: str) -> str:
        """Searches Wikipedia for the given entity and processes the results.

        Args:
            entity: The entity to search for on Wikipedia.

        Functionality:
            - Constructs and sends a search query to Wikipedia.
            - Parses the search results.
            - If similar results are found, returns a list of similar titles.
            - If the entity page is found, returns a summary of the page content.
            - Handles disambiguation pages by recursively searching with modified query.
        """
        if not entity:
            self.state.done = True
            return "Search failed. No entity provided."

        # In case the searched entity is e.g. a year
        search_entity = (
            str(entity) if isinstance(entity, int) else entity.replace(" ", "+")  # type: ignore[redundant-expr,unreachable]  # noqa: FURB123
        )
        try:
            # Access from cache if we previously searched for search_entity
            response_text = self.wiki_cache[search_entity]
        except KeyError:
            # follow_redirects=True because wikipedia frequently redirects to the correct page
            async with httpx.AsyncClient(
                follow_redirects=True, proxy=self.proxy
            ) as client:
                response = await fetch_with_retry(
                    client,
                    f"https://en.wikipedia.org/w/index.php?search={search_entity}",
                    timeout=15,
                )
                response.raise_for_status()  # Raise an HTTPError for bad responses
                # Cache for subsequent tool calls
                self.wiki_cache[search_entity] = response_text = response.text
        soup = BeautifulSoup(response_text, features="html.parser")
        result_divs = soup.find_all("div", {"class": "mw-search-result-heading"})
        if result_divs:  # mismatch
            result_titles = [clean_str(div.get_text().strip()) for div in result_divs]
            self.state.page = None
            return f"Search for {entity!r} returned no results. Similar: {result_titles[:5]}."

        page = [p.get_text().strip() for p in soup.find_all("p") + soup.find_all("ul")]
        if any("may refer to:" in p for p in page):  # Recurse
            return await self.search(entity="[" + entity + "]")
        # Clean up any unicode
        self.state.page = ""
        for p in page:
            if len(p.split(" ")) > 2:  # noqa: PLR2004
                self.state.page += clean_str(p)
                if not p.endswith("\n"):
                    self.state.page += "\n"
        # Extract and concatenate the first five sentences from the provided page content.
        obs_list = [
            s.strip() + "."
            for p in self.state.page.split("\n")
            if p.strip()
            for s in p.split(". ")
            if s.strip()
        ]
        self.state.last_action_is_lookup = False
        return " ".join(obs_list[:5])

    def lookup(self, keyword: str) -> str:
        """Construct a list of sentences from the given page content that contain the specified keyword.

        Args:
            keyword: The keyword to search for within the sentences.

        Returns:
            A list of sentences containing the keyword. If `page` is None or empty, an empty list is returned.

        Example:
            page_content = ("Milhouse Mussolini Van Houten is a recurring character in the Fox animated television series
            The Simpsons voiced by Pamela Hayden and created by Matt Groening. Milhouse is Bart Simpson's best friend in
            Mrs. Krabappel's fourth grade class at Springfield Elementary School. He is an insecure, gullible, and
            less popular child than Bart who is often led into trouble by Bart, who takes advantage of his friend's naivety.
            Milhouse is a regular target for school bully Nelson Muntz and his friends Jimbo Jones, Dolph Starbeam and
            Kearney Zzyzwicz. Milhouse has a crush on Bart's sister, Lisa, a common plot element.")

            keyword = "Milhouse"
            result = lookup(keyword, page_content)
            # Output: ['Milhouse Mussolini Van Houten is a recurring character in the Fox animated television series The
            Simpsons.',
            'Milhouse is Bart Simpson's best friend in Mrs. Krabappel's fourth grade class at Springfield Elementary
            School.',
            'Milhouse is a regular target for school bully Nelson Muntz and his friends Jimbo Jones, Dolph Starbeam and
            Kearney Zzyzwicz.',
            'Milhouse has a crush on Bart's sister, Lisa, a common plot element.']

        """
        if not keyword:
            self.state.done = True
            return "Lookup failed. No keyword provided"

        if not self.state.page:
            return "Lookup failed. You have not specified a Wikipedia page yet."

        if not self.state.last_action_is_lookup or self.state.last_lookup != keyword:
            self.state.last_lookup = keyword
            self.state.lookup_results = [
                s.strip()
                for s in self.state.page.split("\n")
                if s.strip() and keyword.lower() in s.lower()
            ]
            self.state.lookup_index = 0

        if self.state.lookup_index >= len(self.state.lookup_results):
            return "No more results."

        obs = (
            f"(Result {self.state.lookup_index + 1} / {len(self.state.lookup_results)})"
            f" {self.state.lookup_results[self.state.lookup_index]}"
        )
        self.state.lookup_index += 1
        self.state.last_action_is_lookup = True
        return obs


class HotPotQADifficultyLevel(StrEnum):
    EASY = "easy"
    MEDIUM = "medium"
    HARD = "hard"


class HotPotQAEnvConfig(BaseModel):
    """Configuration model for the HotPotQA environment.

    This defines the configuration parameters for setting up the HotPotQA environment,
    including the path to the data repository and the specific path to the prompt file.
    """

    model_config = ConfigDict(extra="forbid")
    PROXY_ENVIRON_VAR_NAME: ClassVar[str] = "AVIARY_ENV_HOTPOTQA_PROXY"

    shuffle_data: bool = Field(
        default=False, description="Set True to shuffle the dataset after loading."
    )
    data_fraction: float = Field(
        default=1.0,
        description=(
            "Fraction of data to use, default is full dataset. If set, data is"
            " subsampled after shuffling (if enabled)."
        ),
        gt=0.0,
        le=1.0,
    )
    correct_reward: float = 1.0
    incorrect_reward: float = 0.0
    tool_failure_reward: float = 0.0
    difficulty_level: set[HotPotQADifficultyLevel] | None = Field(
        None,
        description=(
            "The HotPotQA difficulty level to filter on. "
            "Can be any subset of `{'easy', 'medium', 'hard'}`. "
            "If None, we will not filter any task instances. "
            "Note that not all splits have a 'level' field."
        ),
    )
    proxy: str | None = Field(
        default=None,
        description=(
            "Optional proxy URL to use, default is unused. If unspecified, the"
            f" environment variable {PROXY_ENVIRON_VAR_NAME} will be checked."
        ),
    )

    @field_validator("proxy")
    @classmethod
    def check_proxy_environ_var(cls, value: str | None) -> str | None:
        if not value:
            value = os.environ.get(cls.PROXY_ENVIRON_VAR_NAME, "") or None
        return value

    evaluation_mode: EvalAnswerMode = EvalAnswerMode.CONTAINS


class HotPotQADataset(TaskDataset[HotPotQAEnv]):
    # SEE: https://huggingface.co/datasets/hotpotqa/hotpot_qa
    HOTPOTQA_HUGGING_FACE_DATASET = "hotpotqa/hotpot_qa"

    def get_data_from_hugging_face(
        self, split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
    ) -> list[tuple[str, str]]:
        """Convert a local file and split to a list of (question, answer) tuples."""
        if split in {"dev", "eval", "val"}:  # Map common aliases
            split = "validation"
        all_datasets = load_dataset(hf_dataset, name="fullwiki", trust_remote_code=True)
        try:
            data = all_datasets[split].select_columns(
                column_names=["question", "answer", "level"]
            )
        except KeyError as exc:
            raise ValueError(
                f"Split {split!r} was invalid for Hugging Face dataset {hf_dataset},"
                f" please specify a split from {set(all_datasets.keys())}."
            ) from exc

        if not all(  # Making up for datasets not being typed: https://github.com/huggingface/datasets/issues/3841
            isinstance(d["question"], str) and isinstance(d["answer"], str | float)
            for d in data
        ):
            raise ValueError(
                f"Dataset {hf_dataset!r} and split {split!r} contains invalid"
                " question(s) or answer(s)."
            )
        return [(d["question"], d["answer"]) for d in data if self._filter_task(d)]

    def __init__(
        self, split: str, config: HotPotQAEnvConfig | dict | None = None, **kwargs
    ):
        super().__init__()
        if isinstance(config, dict):  # Serialized config
            config = HotPotQAEnvConfig(**(config | kwargs))
        elif config is None:
            config = HotPotQAEnvConfig(**kwargs)
        self.config = config
        raw_data = self.get_data_from_hugging_face(split)
        if self.config.shuffle_data:
            random.shuffle(raw_data)
        if self.config.data_fraction < 1.0:
            raw_data = raw_data[: int(self.config.data_fraction * len(raw_data))]
        self.data = raw_data

    def _filter_task(self, task: dict[str, Any]) -> bool:
        """Decide whether to keep a task instance based on configuration options."""
        if self.config.difficulty_level is None:
            return True

        try:
            return task["level"] in self.config.difficulty_level
        except KeyError as e:
            raise RuntimeError(
                "Attempting to filter difficulty level, "
                "but this split does not have a 'level' field."
            ) from e

    def get_new_env_by_idx(self, idx: int) -> HotPotQAEnv:
        return HotPotQAEnv(
            *self.data[idx],
            correct_reward=self.config.correct_reward,
            incorrect_reward=self.config.incorrect_reward,
            tool_failure_reward=self.config.tool_failure_reward,
            evaluation_mode=self.config.evaluation_mode,
            proxy=self.config.proxy,
        )

    def __len__(self) -> int:
        return len(self.data)



from __future__ import annotations

import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, ClassVar, Self

from pydantic import BaseModel, Field, field_validator, model_validator

from aviary.utils import encode_image_to_base64

if TYPE_CHECKING:
    from logging import LogRecord

    import numpy as np


class Message(BaseModel):
    DEFAULT_ROLE: ClassVar[str] = "user"
    VALID_ROLES: ClassVar[set[str]] = {
        DEFAULT_ROLE,
        "system",
        "tool",
        "assistant",
        "function",  # Prefer 'tool'
    }

    role: str = Field(
        default=DEFAULT_ROLE,
        description="Message role matching OpenAI's role conventions.",
    )
    content: str | None = Field(
        default=None,
        description=(
            "Optional message content. Can be a string or a dictionary or None. "
            "If a dictionary (for multimodal content), it will be JSON serialized. "
            "None is a sentinel value for the absence of content "
            "(different than empty string)."
        ),
    )
    content_is_json_str: bool = Field(
        default=False,
        description=(
            "Whether the content is JSON-serialized (e.g., for multiple modalities)."
        ),
        exclude=True,
        repr=False,
    )

    info: dict | None = Field(
        default=None,
        description="Optional metadata about the message.",
        exclude=True,
        repr=False,
    )

    @field_validator("role")
    @classmethod
    def check_role(cls, v: str) -> str:
        if v not in cls.VALID_ROLES:
            raise ValueError(f"Role {v} was not in {cls.VALID_ROLES}.")
        return v

    @model_validator(mode="before")
    @classmethod
    def serialize_content(cls, data):
        if not (isinstance(data, dict) and "content" in data):
            return data

        content = data["content"]
        if not content or isinstance(content, str):
            return data

        try:
            data["content"] = json.dumps(content)
            data["content_is_json_str"] = True
        except TypeError as e:
            raise ValueError("Content must be a string or JSON-serializable.") from e

        return data

    def __str__(self) -> str:
        return self.content or ""

    def model_dump(self, *args, **kwargs) -> dict:
        dump = super().model_dump(*args, **kwargs)
        if self.content_is_json_str:
            dump["content"] = json.loads(dump["content"])
        return dump

    def append_text(
        self, text: str, delim: str = "\n", inplace: bool = True
    ) -> Message:
        """Append text to the content.

        Args:
            text: The text to append.
            delim: The delimiter to use when concatenating strings.
            inplace: Whether to modify the message in place.

        Returns:
            The modified message. Note that the original message is modified and returned
            if `inplace=True` and a new message is returned otherwise.
        """
        if not self.content:
            new_content = text
        elif self.content_is_json_str:
            try:
                content_list = json.loads(self.content)
                if not isinstance(content_list, list):
                    raise TypeError("JSON content is not a list.")
                content_list.append({"type": "text", "text": text})
                new_content = json.dumps(content_list)
            except json.JSONDecodeError as e:
                raise ValueError("Content is not valid JSON.") from e
        else:
            new_content = f"{self.content}{delim}{text}"
        if inplace:
            self.content = new_content
            return self
        return self.model_copy(update={"content": new_content}, deep=True)

    @classmethod
    def create_message(
        cls,
        role: str = DEFAULT_ROLE,
        text: str | None = None,
        image: np.ndarray | None = None,
    ) -> Self:
        # Assume no image, and update to image if present
        content: str | list[dict] | None = text
        if image is not None:
            content = [
                {
                    "type": "image_url",
                    "image_url": {"url": encode_image_to_base64(image)},
                }
            ]
            if text is not None:
                content.append({"type": "text", "text": text})
        return cls(role=role, content=content)


def join(
    msgs: Iterable[Message], delimiter: str = "\n", include_roles: bool = True
) -> str:
    return delimiter.join(
        f"{f'{m.role}: ' if include_roles else ''}{m.content or ''}" for m in msgs
    )


class MalformedMessageError(ValueError):
    """Error to throw if some aspect of a Message variant is malformed."""

    @classmethod
    def common_retryable_errors_log_filter(cls, record: LogRecord) -> bool:
        """
        Filter out common parsing failures not worth looking into from logs.

        Returns:
            False if the LogRecord should be filtered out, otherwise True to keep it.
        """
        # NOTE: match both this Exception type's name and its content, to be robust
        return not all(x in record.msg for x in (cls.__name__, EMPTY_CONTENT_BASE_MSG))


class EnvStateMessage(Message):
    """A message that contains the current state of the environment."""


# Define separately so we can filter out this message type
EMPTY_CONTENT_BASE_MSG = "No content in message"



import asyncio
import os

try:
    import click
    import uvicorn
except ImportError as e:
    raise ImportError(
        "CLI requires the 'server' extra for 'aviary'. Please:"
        " `pip install aviary[server]`."
    ) from e

from aviary.env import Environment
from aviary.tools.server import make_tool_server


# this enables sub commands for the CLI
# so we can call `aviary tools` to start the tool server
# rather than aviary-tools or something
@click.group()
def cli():
    pass


@cli.command()
@click.argument("env")
@click.option("--host", default="localhost")
@click.option("--port", default=8000)
@click.option("--token", default="secret")
def tools(env: str, host: str, port: int, token: str):
    if not os.environ.get("AUTH_TOKEN"):
        os.environ["AUTH_TOKEN"] = token

    # use empty task to trigger
    # an empty task/no problem
    def env_factory():
        return Environment.from_name(env, task="")

    app = asyncio.run(make_tool_server(env_factory))
    click.echo(
        f"View tools at http://{host}:{port}/docs and log in with token {token!r}"
    )
    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    cli()



import json
import logging
import os
import shutil
from pathlib import Path
from typing import Any
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator

from aviary.env import Frame

try:
    from boto3 import client
except ImportError:
    client = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class Renderer(BaseModel):
    id: UUID | int | str = Field(
        default_factory=lambda: str(uuid4()).replace("-", "")[:16]
    )
    frames: list[Frame] = []
    prefix: str
    name: str = Field(
        default="Trajectory",
        description="Name of the renderer, used in the manifest file.",
    )

    @field_validator("prefix")
    @classmethod
    def check_prefix_is_alphanum(cls, v: str) -> str:
        if not v.isalnum():
            raise ValueError("Prefix must be an alphanumeric string")
        return v

    def append(self, frame: Frame) -> None:
        self.frames.append(frame)

    def _make_filename(self, index: int) -> str:
        return f"{self.prefix}_{self.id!s}_{index}.json"

    def _render(self) -> dict[str, dict[str, Any]]:
        """Get a mapping of filenames to serialized frame with renderer metadata."""
        return {
            self._make_filename(i): frame.model_dump()
            | {
                "index": i + 1,
                "prev_frame": self._make_filename(i - 1) if i > 0 else None,
                "next_frame": (
                    self._make_filename(i + 1) if i + 1 < len(self.frames) else None
                ),
            }
            for i, frame in enumerate(self.frames)
        }

    def build(
        self,
        build_dir: str | os.PathLike,
        indent: int = 4,
        r2_bucket: str | None = None,
        extra_files: list[str | os.PathLike] | None = None,
    ) -> None:
        name_to_data = self._render()
        if not name_to_data:
            logger.warning("No frames to render.")
            return
        build_dir_path = Path(build_dir)
        file_list = []
        for name, data in name_to_data.items():
            path = build_dir_path / name
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as f:
                json.dump(data, f, indent=indent)
            file_list.append(name)

        # now write manifest (summary) file, and we know name exists
        # because we check name_to_data has values
        # pylint: disable-next=undefined-loop-variable
        first_name, last_name = next(iter(name_to_data.keys())), name
        # NOTE: we have the '-info' prefix so that we can use a prefix-filter
        manifest_fn = f"{self.prefix}_info_{self.id!s}.json"
        with (build_dir_path / manifest_fn).open("w") as f:
            json.dump(
                {"name": self.name, "first": first_name, "last": last_name},
                f,
                indent=indent,
            )
        file_list.append(manifest_fn)

        # copy the extra files
        if extra_files:
            for p in extra_files:
                path = Path(p)
                output_path = build_dir_path / path.name
                output_path.parent.mkdir(parents=True, exist_ok=True)
                try:
                    shutil.copyfile(path, output_path)
                    file_list.append(output_path.name)
                except FileNotFoundError:
                    logger.warning(f"Failed to copy file from {path} to {output_path}.")

        if r2_bucket:  # Upload to bucket
            try:
                CF_ACCOUNT_ID = os.environ["CF_ACCOUNT_ID"]
                CF_ACCESS_KEY_ID = os.environ["CF_ACCESS_KEY_ID"]
                CF_SECRET_ACCESS_KEY = os.environ["CF_SECRET_ACCESS_KEY"]
                if not (CF_ACCOUNT_ID and CF_ACCESS_KEY_ID and CF_SECRET_ACCESS_KEY):
                    raise ValueError("Empty Cloudflare R2 credentials")  # noqa: TRY301
            except (KeyError, ValueError) as exc:
                raise ValueError("Cloudflare R2 credentials unset.") from exc
            try:
                s3 = client(
                    service_name="s3",
                    endpoint_url=f"https://{CF_ACCOUNT_ID}.r2.cloudflarestorage.com",
                    aws_access_key_id=CF_ACCESS_KEY_ID,
                    aws_secret_access_key=CF_SECRET_ACCESS_KEY,
                    region_name="auto",
                )
            except TypeError:
                raise ImportError(
                    "Bucket uploads requires the 'cloud' extra for 'boto3'. Please:"
                    " `pip install aviary[cloud]`."
                ) from None

            for fn in file_list:
                with (build_dir_path / fn).open("rb") as d:
                    s3.upload_fileobj(d, r2_bucket, fn)



import pytest

from aviary.core import eval_answer


@pytest.mark.vcr
@pytest.mark.parametrize(
    ("proposed", "correct", "question", "eval_mode", "expected"),
    [
        pytest.param("\n\n250", "250", None, "exact", True, id="exact"),
        pytest.param(
            "Answer:\n\n250", "250", None, "exact", False, id="exact with noise"
        ),
        pytest.param(
            "Answer\n\n: 250", "250", None, "contains", True, id="contains with noise"
        ),
        pytest.param("A)", "A", None, "contains", True, id="contains multiple choice"),
        pytest.param(
            "The answer is C", "D", None, "contains", False, id="contains wrong answer"
        ),
        pytest.param(
            "Based on all factors considered, the most compelling answer is Gerald, C",
            "C",
            "Which of the following is most likely true:\n\nA) Piggie, B) Pigeon, C) Gerald\n",
            "llm",
            True,
            id="llm basic",
        ),
    ],
)
@pytest.mark.asyncio
async def test_eval_answer(proposed, correct, question, eval_mode, expected):
    assert await eval_answer(proposed, correct, question, eval_mode) == expected


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_eval_llm_config():
    config = {"temperature": 0.5}
    assert await eval_answer("250", "250", "What is 25 * 10?", "llm", config)



import asyncio
import json
import pathlib
import re
import tempfile
import time
from typing import ClassVar

import litellm
import pytest
import pytest_asyncio
from httpx import AsyncClient
from pydantic import BaseModel
from pytest_subtests import SubTests

from aviary.core import (
    DummyEnv,
    DummyEnvState,
    Environment,
    Frame,
    Message,
    Renderer,
    TaskDataset,
    Tool,
    ToolCall,
    ToolRequestMessage,
    ToolResponseMessage,
    ToolsAdapter,
    ToolSelector,
    ToolSelectorLedger,
)
from aviary.dataset_server import TaskDatasetServer
from tests import CILLMModelNames
from tests.conftest import VCR_DEFAULT_MATCH_ON

# Mistral API v0.0.2 required tool calls to comply with this pattern
MISTRAL_API_TOOL_CALL_ID_PATTERN = re.compile(r"^[a-zA-Z0-9]{9}$")


class TestDummyEnv:
    @pytest.mark.asyncio
    async def test_dummyenv(self, dummy_env: DummyEnv) -> None:
        async def my_policy(obs: list[Message]) -> ToolRequestMessage:  # noqa: ARG001
            # For testing purposes, we hardcoded the policy
            return ToolRequestMessage(
                tool_calls=[
                    ToolCall.from_name("print_story", story="Once upon a time done")
                ],
            )

        obs, _ = await dummy_env.reset()
        assert isinstance(obs, list)
        assert len(obs) == 1

        action = await my_policy(obs)
        _, reward, done, _ = await dummy_env.step(action)
        assert reward > 0
        assert done

    @pytest.mark.asyncio
    async def test_tool_signatures(self, dummy_env: DummyEnv) -> None:
        _, tools = await dummy_env.reset()
        assert ToolsAdapter.dump_python(tools, exclude_none=True) == [
            {
                "type": "function",
                "info": {
                    "name": "print_story",
                    "description": "Print a story.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "story": {
                                "type": "string",
                                "title": "Story",
                                "description": "Story to print.",
                            }
                        },
                        "required": ["story"],
                    },
                },
            },
            {
                "info": {
                    "description": "Cast the input argument x to a float.",
                    "name": "cast_float",
                    "parameters": {
                        "properties": {"x": {"type": "string", "title": "X"}},
                        "required": ["x"],
                        "type": "object",
                    },
                },
                "type": "function",
            },
            {
                "info": {
                    "description": "Cast the input argument x to an integer.",
                    "name": "cast_int",
                    "parameters": {
                        "properties": {"x": {"type": "number", "title": "X"}},
                        "required": ["x"],
                        "type": "object",
                    },
                },
                "type": "function",
            },
        ]

    def test_loading_from_name(self):
        env: DummyEnv = Environment.from_name("dummy")
        assert isinstance(env, DummyEnv)

        dataset = TaskDataset.from_name("dummy")
        batch = next(iter(dataset.iter_batches(1)))
        assert len(batch) == 1
        assert isinstance(batch[0], DummyEnv)

    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.OPENAI.value, CILLMModelNames.ANTHROPIC.value]
    )
    @pytest.mark.asyncio
    async def test_tool_calling(self, dummy_env: DummyEnv, model_name: str) -> None:
        def get_todo_list(n: int) -> str:
            """Get todo list for today.

            Args:
                n: number of items to return
            """
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"][:n])

        tool = Tool.from_function(get_todo_list)
        dummy_env.tools = [tool]
        tool_request_message = ToolRequestMessage(
            tool_calls=[ToolCall.from_name("get_todo_list", n=3)]
        )
        assert all(
            MISTRAL_API_TOOL_CALL_ID_PATTERN.match(tc.id)
            for tc in tool_request_message.tool_calls
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        (new_message,) = new_messages
        assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
        assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

        def get_todo_list_no_args() -> str:
            """Get todo list for today."""
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"])

        tool = Tool.from_function(get_todo_list_no_args)
        dummy_env.tools = [tool]
        tool_request_message = ToolRequestMessage(
            tool_calls=[ToolCall.from_name("get_todo_list_no_args")]
        )
        assert all(
            MISTRAL_API_TOOL_CALL_ID_PATTERN.match(tc.id)
            for tc in tool_request_message.tool_calls
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        (new_message,) = new_messages
        assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
        assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

        # ok now try with multiple functions

        def get_calendar() -> str:
            """Get text version of calendar for today."""
            return "9:00am Wake-up\n10:00pm Go to bed\n"

        tool2 = Tool.from_function(get_calendar)
        dummy_env.tools = [tool, tool2]
        tool_request_message = ToolRequestMessage(
            # NOTE: use from_tool to test coverage of that classmethod too
            tool_calls=[ToolCall.from_tool(tool), ToolCall.from_tool(tool2)],
        )
        assert all(
            MISTRAL_API_TOOL_CALL_ID_PATTERN.match(tc.id)
            for tc in tool_request_message.tool_calls
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        if model_name.startswith("claude"):
            # Anthropic not always so smart
            assert 1 <= len(new_messages) <= 2
        else:
            assert len(new_messages) == 2


@pytest.mark.asyncio
async def test_multiple_calls(dummy_env: DummyEnv) -> None:
    obs, tools = await dummy_env.reset()
    calls = [
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
    ]
    action = ToolRequestMessage(tool_calls=calls)
    obs, reward, done, truncated = await dummy_env.step(action)
    assert reward > 0
    assert done


@pytest.mark.parametrize("concurrent_tool_calls", [False, True])
@pytest.mark.asyncio
async def test_invalid_tool_call(
    dummy_env: DummyEnv, concurrent_tool_calls: bool
) -> None:
    def sleep(duration: float) -> None:
        """Sleep for the input duration in seconds."""
        time.sleep(duration)

    sleep_tool = Tool.from_function(sleep, allow_empty_param_descriptions=True)
    _, tools = await dummy_env.reset()
    dummy_env.tools.append(sleep_tool)
    dummy_env.concurrent_tool_calls = concurrent_tool_calls

    obs, *_ = await dummy_env.step(
        ToolRequestMessage(tool_calls=[ToolCall.from_name("invalid_tool")])
    )
    assert obs
    assert obs[0].content
    assert "Invalid tool call" in obs[0].content

    # check that order is preserved even with invalid tool calls
    tool_calls = [
        ToolCall.from_tool(sleep_tool, duration=0.1),
        ToolCall.from_name("invalid_tool"),
        ToolCall.from_name("invalid_tool"),
        ToolCall.from_tool(sleep_tool, duration=0.1),
    ]
    tic = time.perf_counter()
    obs, *_ = await dummy_env.step(ToolRequestMessage(tool_calls=tool_calls))
    if concurrent_tool_calls:
        assert time.perf_counter() - tic < 0.15
    else:
        assert time.perf_counter() - tic > 0.15
    assert obs
    for o, t in zip(obs, tool_calls, strict=True):
        assert o.tool_call_id == t.id


class TestRendering:
    class SomeState(BaseModel):
        field: int

    @pytest.mark.parametrize(
        ("state", "serialized"),
        [
            (5, 5),
            (5.6, 5.6),
            ("hi", "hi"),
            (True, True),
            (["hi"], ["hi"]),
            ({"hi": 5}, {"hi": 5}),
            (SomeState(field=5), {"field": 5}),
        ],
    )
    def test_serialization(self, state, serialized) -> None:
        assert Frame(state=state).model_dump()["state"] == serialized

    def test_frame_mutability(self) -> None:
        # make a nested list - so shallow copy won't catch it
        mutable_state = [["foo"]]
        non_deep_copy = Frame(state=mutable_state, deepcopy=False)
        mutable_state[0].append("bar")
        assert non_deep_copy.model_dump()["state"] == [["foo", "bar"]]

        mutable_state = [["foo"]]
        deep_copy = Frame(state=mutable_state)
        mutable_state[0].append("bar")
        assert deep_copy.model_dump()["state"] == [["foo"]]

    def test_rendering(self, dummy_env: DummyEnv) -> None:
        # Reset to add state
        asyncio.run(dummy_env.reset())

        renderer = Renderer(name="Name", prefix="test")
        renderer.append(dummy_env.export_frame())
        with tempfile.TemporaryDirectory() as tmpdir:
            build_dir = pathlib.Path(tmpdir)
            renderer.build(build_dir)
            file_paths = list(build_dir.glob("*.json"))
            assert len(file_paths) == 2, "Expected manifest and one object"
            frame_path = file_paths[
                file_paths[0].name.removeprefix("test_").startswith("info")
            ]
            with frame_path.open() as f:
                rehydrated = json.load(f)
        assert rehydrated["state"]["messages"] == [
            "Write a 5 word story via print_story"
        ]


class ParallelizedDummyEnv(DummyEnv):
    def __init__(self, right_hand_broken: bool = False):
        super().__init__()
        self.right_hand_broken = right_hand_broken

    RIGHT_HAND_BROKEN_MESSAGE: ClassVar[str] = "Right hand is broken."

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        def move_right_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your right hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            if self.right_hand_broken:  # Use this to test tool errors
                raise RuntimeError(self.RIGHT_HAND_BROKEN_MESSAGE)
            state.reward += 1

        def move_left_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your left hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            state.reward += 1

        def smile_and_wave(state: DummyEnvState) -> None:
            """
            Smile and wave.

            Args:
                state: Current state.
            """
            state.reward = 10
            state.done = True

        self.tools = [
            Tool.from_function(move_left_hand),
            Tool.from_function(move_right_hand),
            Tool.from_function(smile_and_wave),
        ]
        self.state = type(self).State(
            messages=[
                Message(
                    role="user",
                    content=(
                        "You are the president of the United States of America."
                        " Please move both hands at the same time, and then smile"
                        " and wave."
                    ),
                )
            ]
        )
        return self.state.messages, self.tools


class TestParallelism:
    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    async def test_exec_tool_calls_handling(self, model_name: str) -> None:
        env = ParallelizedDummyEnv(right_hand_broken=True)
        obs, tools = await env.reset()
        right_hand_tool = tools[1]

        # 1. Let's DIY create a ToolRequestMessage for test determinism
        request_msg = ToolRequestMessage(
            tool_calls=[ToolCall.from_tool(right_hand_tool, distance=5)]
        )

        # 2. Okay, our hand was broken, let's handle it DIY-style
        try:
            obs, *_ = await env.step(action=request_msg)
        except RuntimeError as exc:
            obs = [
                Message(
                    content=f"Failed to execute tools with message:\n{exc}", role="tool"
                )
            ]
        else:
            raise AssertionError("Should have blown up per the test logic.")

        # 2. Now that we have confirmed that, let's make sure exec_tool_calls
        #    can automate this for us
        obs = await env.exec_tool_calls(
            message=request_msg, state=env.state, handle_tool_exc=True
        )
        (failure_tool_response,) = obs
        assert isinstance(failure_tool_response, ToolResponseMessage)
        assert env.RIGHT_HAND_BROKEN_MESSAGE in failure_tool_response.content

    @pytest.mark.vcr(match_on=[*VCR_DEFAULT_MATCH_ON, "body"])
    @pytest.mark.parametrize("model_name", [CILLMModelNames.OPENAI.value])
    @pytest.mark.asyncio
    async def test_tool_selector_from_model_name(
        self, subtests: SubTests, model_name: str
    ) -> None:
        env = ParallelizedDummyEnv()
        obs, tools = await env.reset()

        with subtests.test("'required' tool_choice"):
            ledger = ToolSelectorLedger(tools=tools)
            selector = ToolSelector(model_name)
            tool_request_message = await selector(obs, tools)
            ledger.messages.append(tool_request_message)
            ledger.model_dump()  # Proving we can serialize the ledger
            assert isinstance(tool_request_message, ToolRequestMessage)
            assert tool_request_message.tool_calls, "Expected at least one tool call"

        with subtests.test("'auto' tool_choice"):
            # NOTE: 'auto' can work, but you risk the ToolSelector not actually
            # selecting a tool, which is why 'auto' is not the default
            ledger = ToolSelectorLedger(tools=tools)
            selector = ToolSelector(model_name)
            tool_request_message = await selector(obs, tools, tool_choice="auto")
            ledger.messages.append(tool_request_message)
            ledger.model_dump()  # Proving we can serialize the ledger
            assert isinstance(tool_request_message, ToolRequestMessage)
            assert tool_request_message.tool_calls, "Expected at least one tool call"

    @pytest.mark.vcr
    @pytest.mark.parametrize("model_name", [CILLMModelNames.OPENAI.value])
    @pytest.mark.asyncio
    async def test_tool_selector_with_external_acompletion(
        self, model_name: str
    ) -> None:
        env = ParallelizedDummyEnv()
        obs_tools = await env.reset()

        router = litellm.Router(
            model_list=[
                litellm.DeploymentTypedDict(
                    model_name="openai", litellm_params={"model": model_name}
                )
            ]
        )
        selector = ToolSelector("openai", router.acompletion)
        tool_request_message = await selector(*obs_tools)
        assert isinstance(tool_request_message, ToolRequestMessage)
        assert tool_request_message.tool_calls, "Expected at least one tool call"

        assert tool_request_message.info, "Expected message info"
        assert tool_request_message.info["usage"][0] > 0, "Expected prompt tokens"
        assert tool_request_message.info["model"], "Expected model name"


@pytest_asyncio.fixture(scope="function")
async def server_async_client() -> AsyncClient:
    dataset = TaskDataset.from_name("dummy")
    server = TaskDatasetServer[DummyEnv](dataset)
    return AsyncClient(app=server.app, base_url="http://test")


class TestTaskDatasetServer:
    @pytest.mark.asyncio
    async def test_start(self, server_async_client: AsyncClient):
        response = await server_async_client.post("/start", json={})
        assert response.status_code == 200
        assert "env_id" in response.json()

    @pytest.mark.asyncio
    async def test_reset_and_step(self, server_async_client: AsyncClient):
        # First, start a new environment
        start_resp = await server_async_client.post("/start", json={})
        env_id = start_resp.json()["env_id"]

        # Now, reset the environment
        response = await server_async_client.post("/reset", json={"env_id": env_id})
        assert response.status_code == 200
        obs, tools = response.json()
        assert isinstance(obs, list)
        assert isinstance(tools, list)

        # Define an action
        action = ToolRequestMessage(
            tool_calls=[
                ToolCall.from_name("print_story", story="Once upon a time done")
            ]
        )

        # Perform a step
        response = await server_async_client.post(
            "/step", json={"env_id": env_id, "action": action.model_dump()}
        )
        assert response.status_code == 200
        obs, reward, done, truncated = response.json()
        assert isinstance(obs, list)
        assert isinstance(reward, float)
        assert isinstance(done, bool)
        assert isinstance(truncated, bool)

    @pytest.mark.asyncio
    async def test_close(self, server_async_client: AsyncClient):
        # Start a new environment
        start_resp = await server_async_client.post("/start", json={})
        env_id = start_resp.json()["env_id"]

        # Close the environment
        response = await server_async_client.post("/close", json={"env_id": env_id})
        assert response.status_code == 200
        assert response.json()["env_id"] == env_id

    @pytest.mark.asyncio
    async def test_close_old_envs(self, server_async_client: AsyncClient):
        # Start a new environment
        await server_async_client.post("/start", json={})

        # Close environments not used in the last 0 seconds
        response = await server_async_client.post(
            "/close_old_envs", json={"last_used": 0}
        )
        assert response.status_code == 200
        assert "closed_env_ids" in response.json()

    @pytest.mark.asyncio
    async def test_info(self, server_async_client: AsyncClient):
        response = await server_async_client.get("/info")
        assert response.status_code == 200
        data = response.json()
        assert data == {
            "dataset_size": None,
            "running_env_ids": [],
        }



from aviary.dataset_server import TaskDatasetServer
from aviary.env import (
    TASK_DATASET_REGISTRY,
    DummyEnv,
    DummyEnvState,
    DummyTaskDataset,
    Environment,
    Frame,
    TaskConfig,
    TaskDataset,
)
from aviary.env_client import (
    EnvironmentClient,
    TaskDatasetClient,
    TaskEnvClientState,
    TaskEnvironmentClient,
)
from aviary.message import EnvStateMessage, MalformedMessageError, Message, join
from aviary.render import Renderer
from aviary.tools import (
    INVALID_TOOL_NAME,
    EvalAnswerMode,
    FunctionInfo,
    Messages,
    MessagesAdapter,
    Parameters,
    Tool,
    ToolCall,
    ToolCallFunction,
    ToolRequestMessage,
    ToolResponseMessage,
    Tools,
    ToolsAdapter,
    ToolSelector,
    ToolSelectorLedger,
    argref_by_name,
    eval_answer,
    wraps_doc_only,
)
from aviary.utils import encode_image_to_base64, is_coroutine_callable, partial_format

__all__ = [
    "INVALID_TOOL_NAME",
    "TASK_DATASET_REGISTRY",
    "DummyEnv",
    "DummyEnvState",
    "DummyTaskDataset",
    "EnvStateMessage",
    "Environment",
    "EnvironmentClient",
    "EvalAnswerMode",
    "Frame",
    "FunctionInfo",
    "MalformedMessageError",
    "Message",
    "Messages",
    "MessagesAdapter",
    "Parameters",
    "Renderer",
    "TaskConfig",
    "TaskDataset",
    "TaskDatasetClient",
    "TaskDatasetServer",
    "TaskEnvClientState",
    "TaskEnvironmentClient",
    "Tool",
    "ToolCall",
    "ToolCallFunction",
    "ToolRequestMessage",
    "ToolResponseMessage",
    "ToolSelector",
    "ToolSelectorLedger",
    "Tools",
    "ToolsAdapter",
    "argref_by_name",
    "encode_image_to_base64",
    "eval_answer",
    "is_coroutine_callable",
    "join",
    "partial_format",
    "wraps_doc_only",
]



import pytest

from aviary.core import Environment, TaskDataset, ToolCall, ToolRequestMessage
from aviary.envs.gsm8k import CalculatorEnv, CalculatorEnvConfig


@pytest.mark.asyncio
async def test_calculator_env() -> None:
    problem = (
        "What is the answer to the ultimate question of life, the universe, and"
        " everything?"
    )
    answer = 42.0
    env = CalculatorEnv(
        problem_id="douglas-adams",
        problem=problem,
        answer=answer,
        config=CalculatorEnvConfig(
            correct_reward=1e4,
        ),
    )

    obs, tools = await env.reset()
    assert obs[0].content == problem
    assert len(tools) == 2

    # Run calculator
    response, reward, done, trunc = await env.step(
        ToolRequestMessage(tool_calls=[ToolCall.from_tool(tools[0], expr="4-3")])
    )
    assert not done
    assert reward == 0.0
    assert response[0].content == "1"

    # check answer
    response, reward, done, trunc = await env.step(
        ToolRequestMessage(tool_calls=[ToolCall.from_tool(tools[1], answer="42")])
    )
    assert reward == 1e4


def test_loading_from_name() -> None:
    env: CalculatorEnv = Environment.from_name(
        "calculator",
        problem_id="rhetorical",
        problem="I had a cake and I ate it. How many cakes do I have?",
        answer=0,
    )
    assert isinstance(env, CalculatorEnv)


@pytest.mark.parametrize(
    ("split", "first_answer"),
    [("train", 10.0), ("train_full", 72.0), ("val", 72.0), ("test", 18.0)],
)
def test_loading_gsm8k_from_name(split: str, first_answer: float) -> None:
    env = TaskDataset.from_name("gsm8k", split=split).get_new_env_by_idx(0)
    assert isinstance(env, CalculatorEnv)
    assert env.answer == first_answer



from .argref import argref_by_name
from .base import (
    INVALID_TOOL_NAME,
    FunctionInfo,
    Messages,
    MessagesAdapter,
    Parameters,
    Tool,
    ToolCall,
    ToolCallFunction,
    ToolRequestMessage,
    ToolResponseMessage,
    Tools,
    ToolsAdapter,
    wraps_doc_only,
)
from .utils import EvalAnswerMode, ToolSelector, ToolSelectorLedger, eval_answer

__all__ = [
    "INVALID_TOOL_NAME",
    "EvalAnswerMode",
    "FunctionInfo",
    "Messages",
    "MessagesAdapter",
    "Parameters",
    "Tool",
    "ToolCall",
    "ToolCallFunction",
    "ToolRequestMessage",
    "ToolResponseMessage",
    "ToolSelector",
    "ToolSelectorLedger",
    "Tools",
    "ToolsAdapter",
    "argref_by_name",
    "eval_answer",
    "wraps_doc_only",
]



import contextlib
import json
from enum import StrEnum
from logging import getLogger
from typing import TYPE_CHECKING, ClassVar, Literal

import datasets
from pydantic import BaseModel, ConfigDict

from aviary.core import (
    Environment,
    Frame,
    Message,
    Messages,
    TaskDataset,
    Tool,
    ToolRequestMessage,
    ToolResponseMessage,
)

if TYPE_CHECKING:
    import pandas as pd

logger = getLogger(__name__)


class CalculatorEnvConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    correct_reward: float = 1.0
    incorrect_reward: float = 0.0
    tool_failure_reward: float = -1.0
    tool_success_reward: float = 0.0
    rel_tol: float = 1e-4

    done_on_failure: bool = True


class CalculatorEnv(Environment[None]):
    def __init__(
        self,
        problem_id: str,
        problem: str,
        answer: float,
        config: CalculatorEnvConfig | None = None,
    ):
        # The problem is not part of the state because it is always the same.
        # Putting it in the state would imply it is somehow affected by .step()
        # or re-initialized by .reset().
        self.problem_id = problem_id
        self.problem = problem
        self.answer = float(answer)  # If passed in as a 0d tensor  # noqa: FURB123

        self.config = config if config is not None else CalculatorEnvConfig()

        self.calc_tool = Tool.from_function(self.calculator)
        self.check_tool = Tool.from_function(self.submit_answer)
        self.tools = [self.calc_tool, self.check_tool]

    @classmethod
    def from_task(cls, task: str) -> "CalculatorEnv":
        return cls(problem_id="task", problem=task, answer=0.0)

    async def reset(self) -> tuple[Messages, list[Tool]]:
        self.state = None  # this environment is effectively stateless
        return [Message(content=self.problem)], self.tools

    async def step(
        self, action: ToolRequestMessage
    ) -> tuple[Messages, float, bool, bool]:
        if not action.tool_calls:
            return (
                [
                    Message(
                        content=(
                            f"Must call one of the provided tools ({self.calculator.__name__} or"
                            f" {self.submit_answer.__name__})."
                        )
                    )
                ],
                self.config.tool_failure_reward,
                self.config.done_on_failure,
                False,
            )

        valid_action, invalid_action = self.filter_invalid_tool_calls(action)

        invalid_response_msgs = [
            ToolResponseMessage.from_call(tool_call, content="")
            for tool_call in invalid_action.tool_calls
        ]

        if valid_action.tool_calls:
            # TODO: Just let exec_tool_calls handle invalid tool calls
            # once someone can take a closer look at what response, reward, done
            # would be in that case.
            results = await self.exec_tool_calls(
                valid_action, handle_invalid_tool_calls=False
            )
            response_msgs = []
            total_reward = 0.0
            any_done = False

            for tool_call, result in zip(valid_action.tool_calls, results, strict=True):
                response, reward, done = json.loads(result.content)

                response_msgs.append(
                    ToolResponseMessage.from_call(tool_call, content=str(response))
                )

                total_reward += reward
                any_done |= done

            return response_msgs + invalid_response_msgs, total_reward, any_done, False

        return (
            invalid_response_msgs,
            self.config.tool_failure_reward * len(invalid_response_msgs),
            self.config.done_on_failure,
            False,
        )

    def submit_answer(self, answer: str) -> tuple[bool, float, Literal[True]]:
        """Submit the proposed answer and check if it is correct. This action is terminal.

        Args:
            answer: Proposed answer.

        Returns:
            Three-tuple of if correct, associated reward (correct_reward if correct,
                tool_failure_reward if tool failure, otherwise incorrect_reward), and
                True indicating done.
        """
        try:
            correct: bool = (
                abs(float(answer) - self.answer)
                / (abs(self.answer) + self.config.rel_tol)
                < self.config.rel_tol
            )
            reward = (
                self.config.correct_reward if correct else self.config.incorrect_reward
            )
        except ValueError:
            return False, self.config.tool_failure_reward, True
        else:
            return correct, reward, True

    def calculator(self, expr: str) -> tuple[float | str | None, float, bool]:
        """Calculate a mathematical expression.

        Args:
            expr (str): A valid python expression

        Returns:
            float: Result of the expression
        """
        try:
            expr = expr.strip()
            result = eval(expr)  # noqa: S307  # pylint: disable=eval-used
            # logger.debug(f"{expr} = {result}")
            with contextlib.suppress(ValueError):
                if int(result) == result:
                    result = int(result)

        except Exception:
            # msg = f"ERROR: ({expr}) -> {repr(e)}"
            # logger.error(msg)
            return (
                "Error using calculator",
                self.config.tool_failure_reward,
                self.config.done_on_failure,
            )
        return result, self.config.tool_success_reward, False

    def export_frame(self) -> Frame:
        return Frame(
            state={
                "problem_id": self.problem_id,
                "problem": self.problem,
                "answer": self.answer,
            }
        )


# SEE: https://huggingface.co/datasets/openai/gsm8k
GSM8K_PUBLIC_SOURCE = "openai/gsm8k"


class GSM8kDatasetSplit(StrEnum):
    train_full = "train_full"  # full training set from OpenAI
    train = "train"  # 80% of train_full (idx%5 != 0)
    val = "val"  # 20% of train_full (idx%5 == 0)
    test = "test"

    def get_df_from_hf(
        self, hf_source: str, add_metadata: bool = True
    ) -> "pd.DataFrame":
        # All non-test splits are derived from train
        hf_split = "test" if self == self.test else "train"

        kw = {}
        if hf_source == GSM8K_PUBLIC_SOURCE:
            kw["name"] = "main"  # as opposed to "socratic"

        src_df = (
            datasets.load_dataset(hf_source, split=hf_split, **kw)
            .to_pandas()
            .reset_index(drop=True)
        )
        if self == self.train:
            src_df = src_df[src_df.index % 5 != 0]
        elif self == self.val:
            src_df = src_df[src_df.index % 5 == 0]
        if add_metadata:
            # Assign problem ID for the env
            src_df["problem_id"] = self.value + "_" + src_df.index.astype(str)

            # Attempt to extract a numerical answer
            try:
                src_df["answer_num"] = src_df["answer"].apply(
                    # answer is formatted as: <some text>\n#### <answer_num>
                    lambda a: float(a.split("#### ")[1].replace(",", ""))
                )
            except Exception as e:
                raise RuntimeError(
                    "Failed to extract numerical answer from 'answer' column"
                ) from e
        return src_df


class GSM8kDataset(TaskDataset):
    Split: ClassVar = GSM8kDatasetSplit

    def __init__(
        self,
        split: GSM8kDatasetSplit | str,
        config: CalculatorEnvConfig | dict | None = None,
        hf_source: str = GSM8K_PUBLIC_SOURCE,
    ):
        if isinstance(config, dict):  # Serialized config
            config = CalculatorEnvConfig(**config)
        elif config is None:
            config = CalculatorEnvConfig()
        self.config = config
        self.src_df = GSM8kDatasetSplit(split).get_df_from_hf(hf_source)

    def get_new_env_by_idx(self, idx: int) -> CalculatorEnv:
        row = self.src_df.iloc[idx]
        return CalculatorEnv(
            problem_id=row["problem_id"],
            problem=row["question"],
            answer=row["answer_num"],
            config=self.config,
        )

    def __len__(self) -> int:
        return len(self.src_df)



import asyncio
import logging
import time
import traceback
import uuid
from contextlib import contextmanager
from itertools import starmap
from typing import Generic, TypeVar

import uvicorn
from pydantic import BaseModel, Field

from aviary.env import Environment, TaskDataset
from aviary.tools import MessagesAdapter, ToolRequestMessage, ToolsAdapter

try:
    from fastapi import FastAPI, HTTPException
except ImportError:
    # We will raise if a TaskDatasetServer is instantiated but fastapi is not available
    FastAPI = HTTPException = None  # type: ignore[misc,assignment]

logger = logging.getLogger(__name__)


class StartRequest(BaseModel):
    task_idx: int | None = Field(
        default=None,
        description="Index of the dataset to start. "
        "If provided, will call TaskDataset.get_new_env_by_idx(); "
        "otherwise, TaskDataset.get_new_env().",
    )


class EnvRequest(BaseModel):
    env_id: str = Field(description="Maps to a running env ID")


class StepRequest(BaseModel):
    env_id: str = Field(description="Maps to a running env ID")
    action: ToolRequestMessage


class FlushRequest(BaseModel):
    last_used: float = Field(default=3600, description="Seconds since last use")


DEFAULT_SERVER_PORT = 8041
BIND_ALL_HOST = "0.0.0.0"  # noqa: S104


# Not sure why, but mypy complains if we use the TEnvironment in aviary.env, so redefine here
TEnvironment = TypeVar("TEnvironment", bound=Environment)


class TaskDatasetServer(Generic[TEnvironment]):
    def __init__(
        self,
        dataset: TaskDataset[TEnvironment],
        host: str = BIND_ALL_HOST,
        port: int = DEFAULT_SERVER_PORT,
    ):
        if FastAPI is None:
            raise ImportError(
                "FastAPI is required to run a TaskDatasetServer. "
                "Please `pip install aviary[server]`."
            )

        self.dataset = dataset
        self.host = host
        self.port = port

        self.app = FastAPI()

        # env ID -> (env, last used timestamp)
        self.envs: dict[str, tuple[TEnvironment, float]] = {}
        self.lock = asyncio.Lock()
        self._setup_routes()

    def _get_env(self, env_id: str) -> TEnvironment:
        try:
            env, _ = self.envs[env_id]
        except KeyError:
            raise HTTPException(
                status_code=404, detail=f"Environment {env_id} not found"
            ) from None
        self.envs[env_id] = (env, time.time())
        return env

    async def close(self):
        await asyncio.gather(*[
            self._close(env_id) for env_id in list(self.envs.keys())
        ])

    async def _close(self, env_id: str):
        try:
            env, _ = self.envs[env_id]
            await env.close()
        except Exception:
            logger.exception(f"Failed to close env {env_id}")

    def _setup_routes(self):
        @self.app.post("/start")
        async def start(req: StartRequest):
            with handle_exc_as_http_exc():
                if req.task_idx is None:
                    env = await asyncio.to_thread(self.dataset.get_new_env)
                else:
                    env = await asyncio.to_thread(
                        self.dataset.get_new_env_by_idx, req.task_idx
                    )

            async with self.lock:
                env_id = str(uuid.uuid4())
                self.envs[env_id] = (env, time.time())
                return {"env_id": env_id}

        @self.app.post("/reset")
        async def reset(req: EnvRequest):
            async with self.lock:
                env = self._get_env(req.env_id)

            with handle_exc_as_http_exc():
                obs, tools = await env.reset()

            return (
                MessagesAdapter.dump_python(obs),
                ToolsAdapter.dump_python(tools, exclude_none=True, by_alias=True),
            )

        @self.app.post("/step")
        async def step(req: StepRequest):
            async with self.lock:
                env = self._get_env(req.env_id)

            with handle_exc_as_http_exc():
                obs, *reward_done_trunc = await env.step(req.action)

            obs_serialized = MessagesAdapter.dump_python(obs)
            return obs_serialized, *reward_done_trunc

        @self.app.post("/close")
        async def close(req: EnvRequest):
            async with self.lock:
                env = self._get_env(req.env_id)
                # Even if env.close() fails, untrack this. It is more likely that
                # the env crashed, not that we somehow lost track of it.
                del self.envs[req.env_id]

                with handle_exc_as_http_exc():
                    await env.close()

            return {"env_id": req.env_id}

        @self.app.post("/close_old_envs")
        async def close_old_envs(req: FlushRequest):
            """Endpoint to close environments that have not been used in a while.

            Useful for cleaning up dangling environments en masse.
            """
            now = time.time()

            async def close(env_id: str, env: TEnvironment) -> str | None:
                try:
                    await env.close()
                except Exception:
                    logger.exception(f"Failed to close env {env_id}")
                    return None
                else:
                    del self.envs[env_id]
                    return env_id

            to_close: list[tuple[str, TEnvironment]] = []
            async with self.lock:
                for env_id, (env, last_used) in list(self.envs.items()):
                    if now - last_used > req.last_used:
                        to_close.append((env_id, env))

                closed = await asyncio.gather(*list(starmap(close, to_close)))

            return {
                "closed_env_ids": [env_id for env_id in closed if env_id is not None]
            }

        @self.app.get("/info")
        def info():
            try:
                dataset_len: int | None = len(self.dataset)
            except TypeError:
                dataset_len = None
            return {
                "dataset_size": dataset_len,
                "running_env_ids": list(self.envs.keys()),
            }

    def start(self):
        uvicorn.run(
            self.app,
            host=self.host,
            port=self.port,
            log_level="debug",
        )

    async def astart(self):
        """Async equivalent of start()."""
        config = uvicorn.Config(
            self.app, host=self.host, port=self.port, log_level="debug"
        )
        server = uvicorn.Server(config)
        await server.serve()


@contextmanager
def handle_exc_as_http_exc():
    # If an environment fails, we don't want to tear down the whole server
    try:
        yield
    except Exception as e:
        raise HTTPException(
            status_code=500, detail=traceback.format_exc() + "\n" + repr(e)
        ) from None



from collections.abc import Callable
from enum import StrEnum
from functools import partial
from typing import TYPE_CHECKING, Any, ClassVar, cast

from pydantic import BaseModel, Field

from aviary.message import MalformedMessageError, Message

from .base import (
    MessagesAdapter,
    Tool,
    ToolRequestMessage,
    ToolResponseMessage,
    ToolsAdapter,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from litellm import ModelResponse


class EvalAnswerMode(StrEnum):
    EXACT = "exact"  # strings must match exactly
    CONTAINS = "contains"  # the correct answer is contained in the supplied answer
    LLM = "llm"  # Ask an LLM to evaluate
    LLM_SCORE = "llm-score"  # Ask an LLM to evaluate and return the score (normalized)


LLM_EVAL_CONFIG = {
    "prompt": (
        "Here is a question, the correct answer to the question, and a proposed answer"
        " to the question. Please tell me if the proposed answer is correct, given the"
        " correct answer. ONLY SAY 'YES' OR 'NO'. No other output is permitted."
        "\n\nQuestion: {question}"
        "\n\nCorrect answer: {correct_answer}"
        "\n\nProposed answer: {proposed_answer}"
    ),
    "model": "gpt-4o-mini",
    "temperature": 0,
}

LLM_SCORE_EVAL_CONFIG = {
    "prompt": (
        "Here is a question, the correct answer to the question, and a rubric for"
        " evaluating the question. Judge the proposed answer based on the given rubric."
        " Give a score from 0 to 10. No other output is permitted."
        "\n\nQuestion: {question}"
        "\n\nRubric: {correct_answer}"
        "\n\nProposed answer: {proposed_answer}"
    ),
    "model": "gpt-4o-mini",
    "temperature": 0,
    "max_score": 10,
}


async def eval_answer(
    proposed: str,
    correct: str,
    question: str | None = None,
    eval_mode: EvalAnswerMode = EvalAnswerMode.CONTAINS,
    llm_eval_config: dict | None = None,
) -> float:
    """Evaluate a proposed answer against a correct answer.

    Will return 0 or 1, except for llm-score which should be between 0 and 1
    """
    if eval_mode in {EvalAnswerMode.LLM, EvalAnswerMode.LLM_SCORE}:
        try:
            from litellm import acompletion
        except ImportError as e:
            raise ImportError(
                "eval_answer requires the 'llm' extra for 'litellm'. Please:"
                " `pip install aviary[llm]`."
            ) from e
        if question is None:
            raise ValueError("Question must be provided for LLM evaluation mode.")
        default_config = (
            LLM_EVAL_CONFIG
            if eval_mode == EvalAnswerMode.LLM
            else LLM_SCORE_EVAL_CONFIG
        )
        config = llm_eval_config or default_config
        prompt = cast(str, config.get("prompt", default_config["prompt"])).format(
            question=question,
            correct_answer=correct,
            proposed_answer=proposed,
        )
        response = await acompletion(
            model=config.get("model", default_config["model"]),
            temperature=config.get("temperature", default_config["temperature"]),
            messages=[{"content": prompt, "role": "user"}],
        )
        if eval_mode == EvalAnswerMode.LLM:
            return await eval_answer(
                response.choices[0].message.content.strip().casefold(),
                "yes",
                eval_mode=EvalAnswerMode.EXACT,
            )
        try:
            return float(response.choices[0].content.strip()) / float(
                config.get("max_score", default_config["max_score"])  # type: ignore[arg-type]
            )
        except ValueError:
            return 0

    gt = correct.strip().casefold()
    pred = proposed.strip().casefold()

    if eval_mode == EvalAnswerMode.EXACT:
        return float(pred == gt)

    if eval_mode == EvalAnswerMode.CONTAINS:
        return float(gt in pred)

    raise RuntimeError(f"Invalid evaluation mode: {eval_mode}")


class ToolSelector:
    """Simple entity to select a tool based on messages."""

    def __init__(
        self,
        model_name: str = "gpt-4o",
        acompletion: "Callable[..., Awaitable[ModelResponse]] | None" = None,
        accum_messages: bool = False,
    ):
        """Initialize.

        Args:
            model_name: Name of the model to select a tool with.
            acompletion: Optional async completion function to use, leaving as the
                default of None will use LiteLLM's acompletion. Alternately, specify
                LiteLLM's Router.acompletion function for centralized rate limiting.
            accum_messages: Whether the selector should accumulate messages in a ledger.
        """
        if acompletion is None:
            try:
                from litellm import acompletion
            except ImportError as e:
                raise ImportError(
                    f"{type(self).__name__} requires the 'llm' extra for 'litellm'."
                    " Please: `pip install aviary[llm]`."
                ) from e
        self._model_name = model_name
        self._bound_acompletion = partial(cast(Callable, acompletion), model_name)
        self._ledger = ToolSelectorLedger() if accum_messages else None

    # SEE: https://platform.openai.com/docs/api-reference/chat/create#chat-create-tool_choice
    # > `required` means the model must call one or more tools.
    TOOL_CHOICE_REQUIRED: ClassVar[str] = "required"

    async def __call__(
        self,
        messages: list[Message],
        tools: list[Tool],
        tool_choice: Tool | str | None = TOOL_CHOICE_REQUIRED,
    ) -> ToolRequestMessage:
        """Run a completion that selects a tool in tools given the messages."""
        completion_kwargs: dict[str, Any] = {}
        # SEE: https://platform.openai.com/docs/guides/function-calling/configuring-function-calling-behavior-using-the-tool_choice-parameter
        expected_finish_reason: set[str] = {"tool_calls"}
        if isinstance(tool_choice, Tool):
            completion_kwargs["tool_choice"] = {
                "type": "function",
                "function": {"name": tool_choice.info.name},
            }
            expected_finish_reason = {"stop"}  # TODO: should this be .add("stop") too?
        elif tool_choice is not None:
            completion_kwargs["tool_choice"] = tool_choice
            if tool_choice == self.TOOL_CHOICE_REQUIRED:
                # Even though docs say it should be just 'stop',
                # in practice 'tool_calls' shows up too
                expected_finish_reason.add("stop")

        if self._ledger is not None:
            self._ledger.messages.extend(messages)
            messages = self._ledger.messages

        model_response = await self._bound_acompletion(
            messages=MessagesAdapter.dump_python(
                messages, exclude_none=True, by_alias=True
            ),
            tools=ToolsAdapter.dump_python(tools, exclude_none=True, by_alias=True),
            **completion_kwargs,
        )

        if (num_choices := len(model_response.choices)) != 1:
            raise MalformedMessageError(
                f"Expected one choice in LiteLLM model response, got {num_choices}"
                f" choices, full response was {model_response}."
            )
        choice = model_response.choices[0]
        if choice.finish_reason not in expected_finish_reason:
            raise MalformedMessageError(
                f"Expected a finish reason in {expected_finish_reason} in LiteLLM"
                f" model response, got finish reason {choice.finish_reason!r}, full"
                f" response was {model_response} and tool choice was {tool_choice}."
            )
        usage = model_response.usage
        selection = ToolRequestMessage(
            **choice.message.model_dump(),
            info={
                "usage": (usage.prompt_tokens, usage.completion_tokens),
                "model": self._model_name,
            },
        )
        if self._ledger is not None:
            self._ledger.messages.append(selection)
        return selection


class ToolSelectorLedger(BaseModel):
    """Simple ledger to record tools and messages."""

    tools: list[Tool] = Field(default_factory=list)
    messages: list[ToolRequestMessage | ToolResponseMessage | Message] = Field(
        default_factory=list
    )



import inspect
import json
import logging
import uuid
from collections.abc import Awaitable, Callable, Iterable
from functools import partial
from itertools import starmap
from typing import Annotated, Any, Literal, NoReturn, Self, TypeAlias

from docstring_parser import DocstringParam, DocstringStyle, parse
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    FieldSerializationInfo,
    PlainSerializer,
    TypeAdapter,
    create_model,
    field_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo

from aviary.message import Message
from aviary.utils import partial_format

try:
    from dicttoxml import dicttoxml
except ImportError:
    dicttoxml = None

logger = logging.getLogger(__name__)

# Mapping from python types to JSON schema types
# SEE: https://json-schema.org/understanding-json-schema/reference/numeric
type_map: dict[type | None, str] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "list",
    dict: "object",
    None: "null",
}
reverse_type_map = {v: k for k, v in type_map.items()}

# A string to denote an invalid tool. It can be used to indicate
# an attempt to use a non-existent tool, missing/invalid parameters,
# mangled output from the LLM, etc.
INVALID_TOOL_NAME = "INVALID"


class ToolCallFunction(BaseModel):
    arguments: dict[str, Any]
    name: str

    @model_validator(mode="before")
    @classmethod
    def deserialize_args(cls, data: Any) -> Any:
        if isinstance(data, dict) and isinstance(data["arguments"], str | None):
            if not data["arguments"]:
                data["arguments"] = {}
            else:
                try:
                    data["arguments"] = json.loads(data["arguments"])
                except json.JSONDecodeError:
                    # If the arguments are not parseable, mark this ToolCall(Function) as invalid
                    # so we can enable "learn"ing what a valid tool call looks like
                    logger.warning(
                        f"Failed to JSON load tool {data.get('name')}'s arguments"
                        f" {data['arguments']}, declaring as {INVALID_TOOL_NAME}."
                    )
                    data["name"] = INVALID_TOOL_NAME
                    data["arguments"] = {}

        return data

    @field_serializer("arguments")
    def serialize_arguments(self, arguments: dict[str, Any]) -> str:
        return json.dumps(arguments)

    def __str__(self) -> str:
        arg_str = ", ".join([f"{k}='{v}'" for k, v in self.arguments.items()])
        return f"{self.name}({arg_str})"


class ToolCall(BaseModel):
    id: str
    type: Literal["function"] = "function"
    function: ToolCallFunction

    @staticmethod
    def generate_id() -> str:
        """Generate a tool call ID of length 9 with values in [a-zA-Z0-9]."""
        return str(uuid.uuid4()).replace("-", "")[:9]

    @classmethod
    def from_tool(cls, tool: "Tool", *args, id: str | None = None, **kwargs) -> Self:  # noqa: A002
        """Create a ToolCall from a Tool and arguments.

        The *args is packaged into the ToolCallFunction's arguments dict with best effort.
        **kwargs is what is passed to toolcall because we have to use named parameters.
        """
        # convert args to kwargs by matching them with the tool's parameters
        for i, name in enumerate(tool.info.parameters.properties.keys()):
            if i < len(args):
                kwargs[name] = args[i]
        return cls(
            id=id or cls.generate_id(),
            function=ToolCallFunction(name=tool.info.name, arguments=kwargs),
        )

    @classmethod
    def from_name(cls, function_name: str, **kwargs) -> Self:
        return cls(
            id=cls.generate_id(),
            function=ToolCallFunction(name=function_name, arguments=kwargs),
        )

    def __str__(self) -> str:
        arg_str = ", ".join([f"{k}='{v}'" for k, v in self.function.arguments.items()])
        return f"{self.function.name}({arg_str})"


class ToolRequestMessage(Message):
    role: Literal["assistant"] = Field(
        default="assistant", description="Matching LiteLLM structure."
    )
    content: str | None = None
    function_call: None = None
    tool_calls: list[ToolCall] = Field(
        default_factory=list,
        description="List of ToolCalls to make concurrently and independently.",
    )

    def __str__(self) -> str:
        if not self.tool_calls:
            return super().__str__()
        base_msg = f"Tool request message {self.content or ''!r}"
        if len(self.tool_calls) == 1:
            return (
                f"{base_msg} for tool calls: "
                f"{self.tool_calls[0]} [id={self.tool_calls[0].id}]"
            )
        return f"{base_msg} for tool calls: " + "; ".join([
            f"{tc!s} [id={tc.id}]" for tc in self.tool_calls
        ])


class ToolResponseMessage(Message):
    content: str = Field(
        description=(
            "Response message content, required to be a string by OpenAI/Anthropic."
        ),
    )
    role: Literal["tool"] = Field(
        default="tool", description="Matching LiteLLM structure."
    )
    name: str = Field(description="Name of the tool that was called.")
    tool_call_id: str = Field(
        description=(
            "Propagated from ToolCall.id, enabling matching response with"
            " ToolRequestMessage."
        )
    )

    @classmethod
    def from_call(cls, call: ToolCall, content: str) -> Self:
        return cls(content=content, name=call.function.name, tool_call_id=call.id)

    @classmethod
    def from_request(
        cls, request: ToolRequestMessage, contents: Iterable[str]
    ) -> list[Self]:
        return list(
            starmap(cls.from_call, zip(request.tool_calls, contents, strict=True))
        )

    def __str__(self) -> str:
        return (
            f"Tool response message {self.content!r} for tool call ID"
            f" {self.tool_call_id} of tool {self.name!r}"
        )


def dict_serialize_exclude_none(
    value: dict[str, dict[str, Any]], info: FieldSerializationInfo
) -> dict[str, dict[str, Any]]:
    """Work around Pydantic not applying exclude_none to dict serializations."""
    if info.exclude_none:
        return {
            p_name: {k: v for k, v in config.items() if v is not None}
            for p_name, config in value.items()
        }
    return value


class Parameters(BaseModel):
    """Matches LiteLLM's desired "tools" schema."""

    model_config = ConfigDict(extra="allow")

    type: Literal["object"] = "object"
    properties: Annotated[
        dict[str, dict[str, Any]], PlainSerializer(dict_serialize_exclude_none)
    ]
    required: list[str]


class FunctionInfo(BaseModel):
    """
    Function-level (not arg-level) information.

    Matches LiteLLM's desired "tools" schema, and resembles inspect.Signature.
    """

    name: str
    description: str
    # SEE: https://github.com/openai/openai-openapi/blob/0f5de60a3d2b263dc2ac362371673f7a21811874/openapi.yaml#L7567-L7570
    parameters: Parameters

    def describe_str(self) -> str:
        for value in self.parameters.properties.values():
            if value.get("allOf") or not value.get("type"):
                raise NotImplementedError(
                    f"Complex types are not yet supported. Failed on: {self!r}"
                )
        # Start with the function prototype
        prototype = f"{self.name}("
        prototype += ", ".join([
            f"{arg['type']} {name}" for name, arg in self.parameters.properties.items()
        ])
        prototype += ")"

        # Function description
        indented_description_lines = "\n".join([
            f"    {line}" if line else "" for line in self.description.split("\n")
        ])
        description = f"DESCRIPTION:\n{indented_description_lines}\n"

        # Parameters description
        params_description = "PARAMETERS:\n"
        for name, arg in self.parameters.properties.items():
            param_desc = (
                f"    {name} ({arg['type']}):"
                f" {arg.get('description') or 'No description provided.'}\n"
            )
            params_description += param_desc

        # Constructing the full man page
        return (
            f"NAME: {self.name}\n\n"
            f"SYNOPSIS:\n    {prototype}\n\n"
            f"{description}\n{params_description}"
        )

    def describe_xml(self) -> str:
        try:
            return dicttoxml(
                self.model_dump(exclude_none=True, by_alias=True),
                custom_root="function_info",
                attr_type=False,
                xml_declaration=False,
            ).decode()
        except TypeError:
            raise ImportError(
                "XML description requires the 'xml' extra for 'dicttoxml'. Please:"
                " `pip install aviary[xml]`."
            ) from None

    def describe_json(self) -> str:
        return self.model_dump_json(exclude_none=True, by_alias=True)

    def __str__(self):
        return self.describe_str()


def _raises(exc: Exception) -> NoReturn:
    """Work around lambda not supporting raise statement."""
    raise exc


class Tool(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    type: Literal["function"] = "function"
    info: FunctionInfo = Field(
        alias="function",
        description=(
            "The serialization alias of 'function' is to match LiteLLM structure on"
            " serialization, and the validation alias enables deserialization."
        ),
    )

    def __init__(
        self,
        tool_fn: Callable[..., Any] | Callable[..., Awaitable[Any]] = (
            lambda *_, **__: _raises(
                NotImplementedError("Please provide a tool function to call.")
            )
        ),
        **kwargs,
    ):
        super().__init__(**kwargs)
        # NOTE: this Callable is excluded from serialization
        self._tool_fn = tool_fn
        self._force_pickle_fn = False

    def __getstate__(self) -> dict[Any, Any]:
        # Prevent _tool_fn from being pickled, SEE: https://stackoverflow.com/a/2345953
        state = super().__getstate__()
        # allow forcing pickle, e.g., for cloud pickle sending
        if self._force_pickle_fn:
            return state
        state["__dict__"] = state["__dict__"].copy()
        state["__dict__"].pop("_tool_fn", None)
        return state

    @staticmethod
    def _get_param_desc(param: DocstringParam, include_type: bool) -> str:
        if not include_type or not param.type_name:
            return param.description or ""
        return f"({param.type_name}): {param.description or ''}"

    @classmethod
    def from_function(
        cls,
        function: Callable[..., Any] | Callable[..., Awaitable[Any]],
        docstring_style: DocstringStyle = DocstringStyle.AUTO,
        allow_empty_param_descriptions: bool = False,
        types_in_param_descriptions: bool = False,
        **formats,
    ) -> "Tool":
        """Hydrate this class via inspection from a free function with a docstring."""
        fxn_name = function.__name__
        # now we parse descriptions from the docstring
        docstring = parse(function.__doc__, style=docstring_style)  # type: ignore[arg-type]  # SEE: https://github.com/rr-/docstring_parser/issues/88
        if not docstring.description:
            raise ValueError(f"Missing docstring for function {fxn_name}.")
        # now we parse descriptions from the docstring
        try:
            # Don't include anything below \f, matching FastAPI's solution for this
            # SEE: https://fastapi.tiangolo.com/advanced/path-operation-advanced-configuration/#advanced-description-from-docstring
            description_stop_index: int | None = docstring.description.index("\\f")
        except ValueError:
            description_stop_index = None
        field_definitions: dict[str, tuple[type, FieldInfo]] = {}
        required: dict[str, bool] = {}
        annotations = function.__annotations__
        for pname, parameter in inspect.signature(function).parameters.items():
            if pname == "state":
                # NOTE: ToolRequestMessage passes state for us, not the LLM
                continue
            d = next(
                (
                    cls._get_param_desc(
                        p, include_type=types_in_param_descriptions
                    ).replace("\n", " ")
                    for p in docstring.params
                    if p.arg_name == pname
                ),
                "",
            )
            if not d and not allow_empty_param_descriptions:
                raise ValueError(f"Missing description for parameter {pname}.")
            required[pname] = parameter.default == inspect.Parameter.empty
            field_config: dict[str, Any] = {}
            if description := partial_format(d, **formats):
                field_config["description"] = description
            if not required[pname]:
                field_config["default"] = parameter.default

            # Annotation resolution order:
            # 1. function.__annotations__: type-hints in function signature or injected
            #    by argref_by_name. If a function has an opinion on a type hint, take it
            #    at face-value.
            # 2. parameter.annotation - this will descend into wrapped functions. For
            #    argref_by_name, this is undesirabe, since the wrapper overwrites type hints.
            #    Hence, this is second in resolution order.
            field_definitions[pname] = (
                annotations.get(pname) or parameter.annotation or type(None),
                Field(**field_config),
            )

        json_schema = create_model(  # type: ignore[call-overload]
            "FieldDefinitions", **field_definitions
        ).model_json_schema()
        json_schema.pop("title")  # Remove the throwaway model name
        if "required" not in json_schema:
            # The API schema doesn't require this, and gpt-3.5-turbo doesn't
            # need this, but claude-3-haiku-20240307 does
            json_schema["required"] = []
        return cls(
            tool_fn=function,
            info=FunctionInfo(
                name=fxn_name,
                description=partial_format(
                    docstring.description[:description_stop_index].strip(), **formats
                ),
                parameters=json_schema,
            ),
        )


def wraps_doc_only(wrapped):
    """A decorator to copy only the docstring from the wrapped function.

    You cannot use functools wraps directly because it will set the __wrapped__ attribute,
    which causes inspect.signature to inspect the wrapped function instead of the wrapper.

    Usage:
        def my_documented_function(foo):
            '''This is a function that does something with foo.'''
            pass

        @wraps_doc_only(my_documented_function)
        def my_other_function(foo, state):
            pass

    In this example, the second function can have different arguments, types, etc. and only the docstring
    will be copied over.
    """

    def _wraps_doc_only(wrapper, wrapped):
        wrapper.__doc__ = wrapped.__doc__
        return wrapper

    return partial(_wraps_doc_only, wrapped=wrapped)


# Conveniences for deserialization
Messages: TypeAlias = list[ToolRequestMessage | ToolResponseMessage | Message]
MessagesAdapter = TypeAdapter(Messages)
Tools: TypeAlias = list[Tool]
ToolsAdapter = TypeAdapter(Tools)



import json

import numpy as np
import pytest

from aviary.core import (
    Message,
    ToolCall,
    ToolCallFunction,
    ToolRequestMessage,
    ToolResponseMessage,
)


class TestMessage:
    def test_roles(self) -> None:
        # make sure it rejects invalid roles
        with pytest.raises(ValueError):  # noqa: PT011
            Message(role="invalid", content="Hello, how are you?")
        # make sure it accepts valid roles
        Message(role="system", content="Respond with single words.")

    @pytest.mark.parametrize(
        ("message", "expected"),
        [
            (Message(), ""),
            (Message(content="stub"), "stub"),
            (Message(role="system", content="stub"), "stub"),
            (ToolRequestMessage(), ""),
            (ToolRequestMessage(content="stub"), "stub"),
            (
                ToolRequestMessage(
                    content="stub",
                    tool_calls=[
                        ToolCall(
                            id="1",
                            function=ToolCallFunction(name="name", arguments={"hi": 5}),
                        )
                    ],
                ),
                "Tool request message 'stub' for tool calls: name(hi='5') [id=1]",
            ),
            (
                ToolRequestMessage(
                    tool_calls=[
                        ToolCall(
                            id="1",
                            function=ToolCallFunction(name="foo1", arguments={"hi": 5}),
                        ),
                        ToolCall(
                            id="2",
                            function=ToolCallFunction(name="foo2", arguments={}),
                        ),
                        ToolCall(
                            id="3",
                            function=ToolCallFunction(name="foo1", arguments=""),
                        ),
                        ToolCall(
                            id="4",
                            function=ToolCallFunction(name="foo2", arguments=None),
                        ),
                    ],
                ),
                (
                    "Tool request message '' for tool calls: "
                    "foo1(hi='5') [id=1]; foo2() [id=2]; foo1() [id=3]; foo2() [id=4]"
                ),
            ),
            (
                ToolResponseMessage(content="stub", name="name", tool_call_id="1"),
                "Tool response message 'stub' for tool call ID 1 of tool 'name'",
            ),
            (
                Message(
                    content=[
                        {"type": "text", "text": "stub"},
                        {"type": "image_url", "image_url": {"url": "stub_url"}},
                    ]
                ),
                (
                    '[{"type": "text", "text": "stub"}, {"type": "image_url",'
                    ' "image_url": {"url": "stub_url"}}]'
                ),
            ),
        ],
    )
    def test_str(self, message: Message, expected: str) -> None:
        assert str(message) == expected

    @pytest.mark.parametrize(
        ("message", "expected"),
        [
            (Message(), {"role": "user"}),
            (Message(content="stub"), {"role": "user", "content": "stub"}),
            (
                Message(
                    content=[
                        {"type": "text", "text": "stub"},
                        {"type": "image_url", "image_url": {"url": "stub_url"}},
                    ]
                ),
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": "stub"},
                        {"type": "image_url", "image_url": {"url": "stub_url"}},
                    ],
                },
            ),
        ],
    )
    def test_dump(self, message: Message, expected: dict) -> None:
        assert message.model_dump(exclude_none=True) == expected

    def test_image_message(self) -> None:
        # An RGB image of a red square
        image = np.zeros((32, 32, 3), dtype=np.uint8)
        image[:] = [255, 0, 0]  # (255 red, 0 green, 0 blue) is maximum red in RGB
        message_text = "What color is this square? Respond only with the color name."
        message_with_image = Message.create_message(text=message_text, image=image)
        assert message_with_image.content
        specialized_content = json.loads(message_with_image.content)
        assert len(specialized_content) == 2
        text_idx, image_idx = (
            (0, 1) if specialized_content[0]["type"] == "text" else (1, 0)
        )
        assert specialized_content[text_idx]["text"] == message_text
        assert "image_url" in specialized_content[image_idx]


class TestToolRequestMessage:
    def test_from_request(self) -> None:
        trm = ToolRequestMessage(
            content="stub",
            tool_calls=[
                ToolCall(
                    id="1",
                    function=ToolCallFunction(name="name1", arguments={"hi": 5}),
                ),
                ToolCall(id="2", function=ToolCallFunction(name="name2", arguments={})),
            ],
        )
        assert ToolResponseMessage.from_request(trm, ("stub1", "stub2")) == [
            ToolResponseMessage(content="stub1", name="name1", tool_call_id="1"),
            ToolResponseMessage(content="stub2", name="name2", tool_call_id="2"),
        ]

    def test_append_text(self) -> None:
        trm = ToolRequestMessage(
            content="stub", tool_calls=[ToolCall.from_name("stub_name")]
        )
        trm_inplace = trm.append_text("text")
        assert trm.content == trm_inplace.content == "stub\ntext"
        # Check append performs an in-place change by default
        assert trm.tool_calls[0] is trm_inplace.tool_calls[0]

        trm_copy = trm.append_text("text", inplace=False)
        assert trm_copy.content == "stub\ntext\ntext"
        # Check append performs a deep copy when not inplace
        assert trm.content == "stub\ntext"
        assert trm.tool_calls[0] is not trm_copy.tool_calls[0]
